unit dmraid;

interface

uses classes, types, sysutils, vv;

type disk_descriptor = record
   number   : U32;
   major    : U32;
   minor    : U32;
   raid_disk   : U32;
   state       : U32;
   padding     : array[1..27] of U32;
end;

type raid_superblock = record
   md_magic       : U32;
   major_version  : U32;
   minor_version  : U32;
   patch_version  : U32;
   gvalid_words   : U32;
   set_uuid0      : U32;
   ctime          : U32;
   level          : S32;
   size           : U32;
   nr_disks       : U32;
   raid_disks     : U32;
   md_minor       : U32;
   not_persistent : U32;
   set_uuid1      : U32;
   set_uuid2      : U32;
   set_uuid3      : U32;
   padding        : array[1..16] of U32;

   // generic
   utime          : U32;
   state          : U32;
   active_disks   : U32;
   working_disks  : U32;
   failed_disks   : U32;
   spare_disks    : U32;
   sb_csum        : U32;
   events_hi      : U32;
   events_lo      : U32;
   cp_events_hi   : U32;
   cp_events_lo   : U32;
   recovery_cp    : U32;
   reshape_position : U32;
   new_level      : U32;
   delta_disks    : U32;
   new_layout     : U32;
   new_chunk      : U32;
   padding2       : array[1..15] of U32;

   // personality
   layout         : U32;
   chunk_size     : U32;
   root_pv        : U32;
   root_block     : U32;
   padding3       : array[1..60] of U32;

   // disks
   disk           : array[0..26] of disk_descriptor;
   //padding[4]     : array[1..?] of U32;

   this_disk      : disk_descriptor;
end;

Praid_superblock = ^raid_superblock;

type TRaidSet = class(TObject)
   level : S32;
   UID   : String;
   Disks : TStringList; // (Superblock, Volume)

   constructor Create;
   destructor Destroy; override;

   function Assemble : TVirtualVolume;
   procedure SortDisks;
   function BitSet(State : U32; Bit : Integer) : Boolean;
end;

function RaidMetadataOffset(Size : Int64) : Int64;
function AssembleRaid(RaidVolumes : TList; VolumeList : TList) : Boolean;

  type TRaidVolume = class(TBoundryVolume)
  private
    Disks      : TList;
    DataDisks  : Integer;
    Size       : Int64;
    ChunkSize  : Integer;
    Algorithm  : Integer;
  public
    constructor Create; override;
    destructor Destroy; override;

    function BoundryRead(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;
    function BoundryWrite(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;

    function Open(WriteAccess : Boolean) : Boolean; override;
    function Close : Boolean; override;

    function GetLength : Int64; override;


    function Start(Super : Praid_superblock; Volumes : TStringList) : Boolean;

  end;



const
 MD_SB_MAGIC = $a92b4efc;
 ALG_RAID4 = 40;
 ALG_RAID0 = 100;

 // array state flag bits
 MD_SB_CLEAN           =  0;
 MD_SB_ERRORS          =  1;

 // disk state flag bits
 MD_DISK_FAULTY        =  0; //* disk is faulty / operational */
 MD_DISK_ACTIVE        =  1; //* disk is running or spare disk */
 MD_DISK_SYNC          =  2; //* disk is in sync with the raid set */
 MD_DISK_REMOVED       =  3; //* disk is in sync with the raid set */



implementation

uses debug, vv_map;

const Facility = 'raid';
              
constructor TRaidSet.Create;
begin
   inherited;

   Disks := TStringList.Create;
end;

destructor TRaidSet.Destroy;
begin
   Disks.Free;

   inherited;
end;

procedure TRaidSet.SortDisks;
var
   NewDisks    : TStringList;
   i           : Integer;
   DiskSuper   : Praid_superblock;
begin
   NewDisks := TStringList.Create;

   for i := 0 to Disks.Count - 1 do
   begin
      NewDisks.Add('');
   end;

   for i := 0 to Disks.Count - 1 do
   begin
      DiskSuper := Pointer(Disks[i]);
      Log('We have disk number ' + IntToStr(DiskSuper.this_disk.number), Facility, LOG_VERBOSE);
      // what do we do if the disk number is too large?
      if DiskSuper.this_disk.number < Disks.Count then
      begin
         NewDisks[DiskSuper.this_disk.number] := Disks[i];
         NewDisks.Objects[DiskSuper.this_disk.number] := Disks.Objects[i];
      end
      else
      begin
         Log('Disk number is too large', Facility, LOG_NORMAL);
      end;
   end;
   Disks.Free;
   Disks := NewDisks;
end;

function TRaidSet.BitSet(State : U32; Bit : Integer) : Boolean;
begin
   if (State and (1 shl Bit)) = 0 then
   begin
      Result := False;
   end
   else
   begin
      Result := True;
   end;
end;


function TRaidSet.Assemble : TVirtualVolume;
var
   Super       : Praid_superblock;
   DiskSuper   : Praid_superblock;
   ddsks : U32;
   v     : TVirtualVolume;
   n     : TVirtualVolume;
   i     : Integer;
   Size  : Int64;
   utime : U32;

begin
   Result := nil;

   Super := Pointer(Disks[0]);

   ddsks := 0;
   if level = 0 then
   begin
      ddsks := Super.nr_disks;
   end
   else if level = 1 then
   begin
      ddsks := 1;
   end
   else if level = 5 then
   begin
      ddsks := Super.nr_disks - 1;
   end;
   Log('Device size ' + IntToStr(Super.size), Facility, LOG_VERBOSE);
   Log('Number of disks ' + IntToStr(Super.nr_disks), Facility, LOG_VERBOSE);
   Log('Number of raid disks ' + IntToStr(Super.raid_disks), Facility, LOG_VERBOSE);
   Log('Number of spare disks ' + IntToStr(Super.spare_disks), Facility, LOG_VERBOSE);
   Log('Array size ' + IntToStr(ddsks * Super.size), Facility, LOG_VERBOSE);
   Log('Chunk size ' + IntToStr(Super.chunk_size), Facility, LOG_VERBOSE);
   utime := Super.utime;

   // sanity check all the disks?
   for i := 0 to Disks.Count - 1 do
   begin
      DiskSuper := Pointer(PChar(Disks[i]));
      if DiskSuper.utime <> utime then
      begin
         Log('utime missmatch');
      end;
      if BitSet(DiskSuper.this_disk.state, MD_DISK_FAULTY) then
      begin
         Log('Disk is marked as FAULTY');
      end;
      if BitSet(DiskSuper.this_disk.state, MD_DISK_ACTIVE) then
      begin
         Log('Disk is marked as ACTIVE');
      end;
      if BitSet(DiskSuper.this_disk.state, MD_DISK_SYNC) then
      begin
         Log('Disk is marked as SYNC');
      end;
      if BitSet(DiskSuper.this_disk.state, MD_DISK_REMOVED) then
      begin
         Log('Disk is marked as REMOVED');
      end;
      v := TVirtualVolume(Disks.Objects[i]);
      if v.GetLayer(facility) <> LEVEL_PROCESSED then
      begin
         if Level >= 0 then
         begin
            v.AddName('RAID' + IntToStr(level) + '/' + UID  + ' disk ' + IntToStr(DiskSuper.this_disk.number));
         end
         else
         begin
            v.AddName('LINEAR/' + UID  + ' disk ' + IntToStr(DiskSuper.this_disk.number));
         end;
         v.SetLayer(Facility, LEVEL_PROCESSED);
      end;
   end;

   if level = 0 then
   begin
      if Disks.Count = Super.nr_disks then
      begin
         n := TRaidVolume.Create;
         n.AddName('RAID0/' + UID);
         if TRaidVolume(n).Start(Super, Disks) then
         begin
            Result := n;
         end;
      end;
   end
   else if level = 1 then
   begin
      // just use the first disk (read only)
      if Disks.Count = Super.nr_disks then
      begin
         v := TVirtualVolume(Disks.Objects[0]);
         n := TMapVolume.Create;
         with n as TMapVolume do
         begin
            AddName('RAID1/' + UID);
            Size := Super.size;
            Size := Size * 1024;
            AddMap(0, Size, v, 0);
         end;
         Result := n;
      end;
   end
   else if level = 4 then
   begin
      if Disks.Count = Super.nr_disks then
      begin
         n := TRaidVolume.Create;
         n.AddName('RAID4/' + UID);
         if TRaidVolume(n).Start(Super, Disks) then
         begin
            Result := n;
         end;
      end;
   end
   else if level = 5 then
   begin
      if Disks.Count = Super.nr_disks then
      begin
         SortDisks;
         n := TRaidVolume.Create;
         n.AddName('RAID5/' + UID);
         if TRaidVolume(n).Start(Super, Disks) then
         begin
            Result := n;
         end;
      end;
   end
   else if Level = -1 then
   begin
      if Disks.Count = Super.nr_disks then
      begin
         SortDisks;
      // we need a loop
      // and to sort the disks by number
         n := TMapVolume.Create;
         n.AddName('LINEAR/' + UID);

         for i := 0 to Disks.Count - 1 do
         begin
            v := TVirtualVolume(Disks.Objects[i]);
            with n as TMapVolume do
            begin
               Size := Super.size;
               Size := Size * 1024;
               AddMap(i * Size, Size, v, 0);
            end;
         end;

         Result := n;
      end;
   end;

   if Assigned(Result) then
   begin
      for i := 0 to Disks.Count - 1 do
      begin
         v := TVirtualVolume(Disks.Objects[i]);
         v.SetLayer(Facility, LEVEL_ACTIVATED);
      end;
   end;
end;

function RaidMetadataOffset(Size : Int64) : Int64;
begin
   //y = (x & ~(MD_RESERVED_BYTES - 1)) - MD_RESERVED_BYTES
   Result := Size or $ffff;
   Result := Result xor $ffff;
   Result := Result - $10000;
end;

function AssembleRaid(RaidVolumes : TList; VolumeList : TList) : Boolean;
var
   Buffer   : String;
   Offset   : Int64;
   Super    : Praid_superblock;
   UID      : String;
   i        : Integer;
   r        : Integer;
   v        : TVirtualVolume;
   Supported : Boolean;

   RaidSets : TList;
   RaidSet  : TRaidSet;
begin
   Result := False;
   // read the super blocks
   // group into raidsets
   // build map volumes
   // add to VolumeList

   RaidSets := TList.Create;

   for i := 0 to RaidVolumes.Count - 1 do
   begin
      v := TVirtualVolume(RaidVolumes[i]);

      SetLength(Buffer, 4096);
      Offset := RaidMetadataOffset(v.GetLength);
      if v.Read(Offset, Length(Buffer), PChar(Buffer)) then
      begin
         Super := Pointer(Buffer);
         Log('magic = ' + IntToHex(Super.md_magic, 8), Facility, LOG_VERBOSE);
         if Super.major_version > 0 then
         begin
            Log('Version ' + IntToStr(Super.major_version) + ' is not supported', Facility, LOG_VERBOSE);
            continue;
         end;
         Log('version = ' + IntToStr(Super.major_version) + '.' + IntToStr(Super.minor_version) + '.' + IntToStr(Super.patch_version), Facility, LOG_VERBOSE);
         Log('level = ' + IntToStr(Super.level), Facility, LOG_VERBOSE);
         Supported := False;
         if Super.level = 1 then
         begin
//            Log(' Mirroring is easy', Facility, LOG_VERBOSE);
            Supported := True;
         end
         else if Super.level = 0 then
         begin
//            Log(' striping is doable', Facility, LOG_VERBOSE);
            Supported := True;
         end
         else if Super.level = 4 then
         begin
//            Log(' RAID4 is WIP', Facility, LOG_VERBOSE);
            Supported := True;
         end
         else if Super.level = 5 then
         begin
//            Log(' RAID5 is WIP', Facility, LOG_VERBOSE);
            Supported := True;
         end
         else if Super.level = -1 then
         begin
//            Log(' LINEAR is easy', Facility, LOG_VERBOSE);
            Supported := True;
         end
         else
         begin
            Log(' Unsupported raid level', Facility, LOG_VERBOSE);
         end;

         if Supported then
         begin
            UID := IntToHex(Super.set_uuid0, 8) + ':' + IntToHex(Super.set_uuid1, 8) + ':' + IntToHex(Super.set_uuid2, 8) + ':' + IntToHex(Super.set_uuid3, 8);
            Log(' We need ' + UID, Facility, LOG_VERBOSE);

            RaidSet := nil;
            for r := 0 to RaidSets.Count - 1 do
            begin
               RaidSet := TRaidSet(RaidSets[r]);
               if RaidSet.UID = UID then
               begin
                  break;
               end
               else
               begin
                  RaidSet := nil;
               end;
            end;

            if Assigned(RaidSet) then
            begin
               if RaidSet.level <> Super.level then
               begin
                  Log('RAID level missmatch', Facility, LOG_ERROR);
               end
               else
               begin
                  RaidSet.Disks.AddObject(Buffer, v);
               end;
            end
            else
            begin
               RaidSet := TRaidSet.Create;
               RaidSet.level := Super.level;
               RaidSet.UID := UID;
               RaidSet.Disks.AddObject(Buffer, v);
               RaidSets.Add(RaidSet);
            end;
         end;
      end;

   end;

   for r := 0 to RaidSets.Count - 1 do
   begin
      RaidSet := TRaidSet(RaidSets[r]);
      v := RaidSet.Assemble;
      if Assigned(v) then
      begin
         VolumeList.Add(v);
         Result := True;
      end;
   end;

   RaidSets.Free;
end;

////////////////////////////////////////////////////////////////////////////////

constructor TRaidVolume.Create;
begin
   inherited;

   Disks := TList.Create;
end;

destructor TRaidVolume.Destroy;
begin
   Disks.Free;

   inherited;
end;

function TRaidVolume.Start(Super : Praid_superblock; Volumes : TStringList) : Boolean;
var
   i : Integer;
   DiskSuper : Praid_superblock;
begin
   Size := (Super.nr_disks - 1) * Super.size * 1024;
   ChunkSize := Super.chunk_size;

   SetBoundry(ChunkSize);

   for i := 0 to Super.nr_disks - 1 do
   begin
      Disks.Add(Volumes.Objects[i]);
   end;

{   for i := 0 to Super.nr_disks - 1 do
   begin
      DiskSuper := Pointer(Volumes[i]);
      Log('We have disk number ' + IntToStr(DiskSuper.this_disk.number), Facility, LOG_VERBOSE);
      Disks[DiskSuper.this_disk.number] := Volumes.Objects[i];
   end;}

   if Super.level = 5 then
   begin
      Algorithm := Super.layout;
      DataDisks := Disks.Count - 1;
   end
   else if Super.level = 4 then
   begin
      Algorithm := ALG_RAID4;
      DataDisks := Disks.Count - 1;
   end
   else if Super.level = 0 then
   begin
      Algorithm := ALG_RAID0;
      DataDisks := Disks.Count;
   end;

   Log('Using layout ' + IntToStr(Algorithm), Facility, LOG_VERBOSE);
   if Algorithm = 0 then
   begin
      Log('left-asymmetric', Facility, LOG_VERBOSE);
      Result := True;
   end
   else if Algorithm = 1 then
   begin
      Log('right-asymmetric', Facility, LOG_VERBOSE);
      Result := True;
   end
   else if Algorithm = 2 then
   begin
      Log('left-symmetric (default)', Facility, LOG_VERBOSE);
      Result := True;
   end
   else if Algorithm = 3 then
   begin
      Log('right-symmetric', Facility, LOG_VERBOSE);
      Result := True;
   end
   else if Algorithm = ALG_RAID4 then
   begin
      Log('RAID4', Facility, LOG_VERBOSE);
      Result := True;
   end
   else if Algorithm = ALG_RAID0 then
   begin
      Log('RAID0', Facility, LOG_VERBOSE);
      Result := True;
   end
   else
   begin
      Log('Unknown algorithm', Facility, LOG_ERROR);
      Result := False;
   end;

   // testing
{   Algorithm := 3;
   Disks.Add(0);
   Disks.Add(0);
   for i := 0 to 23 do
   begin
      Read(i * ChunkSize, 0, nil);
   end;}

end;

// ignore the parity for now.
// we need to check the algorithm so we know which disks have the data
// for each block

// Raid 4 is like raid 5 but the parity disk is always the same
// Raid 0 is like raid 5 but there is no parity disk.

// http://www.accs.com/p_and_p/RAID/LinuxRAID.html
// raid5.c:raid5_compute_sector()

// We could use a map volume, but the large number of chunks means
// that we would use up too much memory.
// We could make a callback function which made the maps up on demand
// but that is still a bit backwards and complex

function TRaidVolume.BoundryRead(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean;
var
   Chunk       : Int64;
   Disk        : Integer;
   ChunkOffset : Int64;

   Stripe      : Integer;
   Parity      : Integer;
begin
   Chunk := Offset div ChunkSize;
   Log('Read chunk     ' + IntToStr(Chunk), Facility, LOG_MAX);
   ChunkOffset := Offset mod ChunkSize;

   if ChunkOffset + Length > ChunkSize then
   begin
      Log('Read is too big for RAID chunk', Facility, LOG_ERROR);
      Result := False;
      exit;
   end;

   Stripe := Chunk div DataDisks;
   // work out which disk and which chunk on that disk
   if Algorithm = 0 then // left a
   begin
      Parity := Disks.Count - (Stripe mod Disks.Count) - 1; // left
      Disk := Chunk mod DataDisks;
      if Disk >= Parity then
      begin
         Disk := Disk + 1;
      end;
   end
   else if Algorithm = 1 then // right a
   begin
      Parity := Stripe mod Disks.Count; // right
      Disk := Chunk mod DataDisks;
      if Disk >= Parity then
      begin
         Disk := Disk + 1;
      end;
   end
   else if Algorithm = 2 then // left s
   begin
      Parity := Disks.Count - (Stripe mod Disks.Count) - 1; // left
      Disk := Chunk mod DataDisks;
      Disk := Disk + Parity + 1;
      Disk := Disk mod Disks.Count;
   end
   else if Algorithm = 3 then // right s
   begin
      Parity := Stripe mod Disks.Count; // right
      Disk := Chunk mod DataDisks;
      Disk := Disk + Parity + 1;
      Disk := Disk mod Disks.Count;
   end
   else if Algorithm = ALG_RAID4 then // raid 4
   begin
      Parity := Disks.Count;
      Disk := Chunk mod DataDisks;
   end
   else if Algorithm = ALG_RAID0 then // raid 0
   begin
      Parity := -1;
      Disk := Chunk mod DataDisks;
   end
   else
   begin
      Log('Invalid RAID algorithm', Facility, LOG_ERROR);
      Result := False;
      exit;
   end;
   Log('Stripe         ' + IntToStr(Stripe), Facility, LOG_MAX);
   Log('Parity disk is ' + IntToStr(Parity), Facility, LOG_MAX);
   Log('Data disk is   ' + IntToStr(Disk), Facility, LOG_MAX);

   Log('', Facility, LOG_MAX);

   Result := TVirtualVolume(Disks[Disk]).Read(Stripe * ChunkSize + ChunkOffset, Length, Buffer);
end;

function TRaidVolume.BoundryWrite(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean;
begin
   Result := False;
end;

function TRaidVolume.Open(WriteAccess : Boolean) : Boolean;
begin
   Result := True;
end;

function TRaidVolume.Close : Boolean;
begin
   Result := True;
end;

function TRaidVolume.GetLength : Int64;
begin
   Result := Size;
end;

////////////////////////////////////////////////////////////////////////////////

function BuildRaidMaps(VolumeList : TList; Options : TStringList) : Boolean;
var
   i : Integer;
   v : TVirtualVolume;
   RaidVolumes : TList;

   Buffer      : String;
   Magic       : PU32;
   Offset      : Int64;

begin
   Result := False;
   RaidVolumes := TList.Create;

   for i := 0 to VolumeList.Count - 1 do
   begin
      v := TVirtualVolume(VolumeList[i]);
      if v.GetLayer(Facility) = LEVEL_ACTIVATED then
      begin
         // Nothing to do
      end
      else if v.GetLayer(Facility) = LEVEL_PROCESSED then
      begin
         // We know it is an unassembeled memeber disk
         RaidVolumes.Add(v);
      end
      else if v.GetLayer(Facility) = LEVEL_NONE then
      begin
         // check this disk
         v.SetLayer(Facility, LEVEL_QUERIED);
         SetLength(Buffer, 512);
         Offset := v.GetLength;
         if Offset > 65536 then
         begin
            Offset := RaidMetadataOffset(Offset);
            Log('Looking for RAID at ' + IntToStr(Offset), Facility, LOG_LOOP);
            if v.Read(Offset, 512, PChar(Buffer)) then
            begin
               Magic := Pointer(Buffer);
               if Magic^ = MD_SB_MAGIC then
               begin
                  Log('Linux MD RAID found. magic = ' + IntToHex(Magic^, 8), Facility, LOG_VERBOSE);

                  RaidVolumes.Add(v);
               end
            end;
         end;
      end;
   end;
   if RaidVolumes.Count > 0 then
   begin
      if AssembleRaid(RaidVolumes, VolumeList) then
      begin
         Result := True;
      end;
   end;
   RaidVolumes.Free;
end;

initialization
   RegisterVolumeScan(Facility, BuildRaidMaps);

end.
