unit execute;

interface


uses
  Windows, Dialogs, Sysutils, Forms, Controls;

type
  TChildProcess = class
  public
   pi                : TProcessInformation;
   hPipeOutputRead   : THandle;
   hPipeInputWrite   : THandle;

   function StartChild(Command : String) : Boolean;
   procedure CloseChild;
   function Read(Buffer : PChar; Length : Integer; Done : PInteger) : Boolean;
   function Write(Buffer : PChar; Length : Integer; Done : PInteger) : Boolean;

   function ReadFull(Buffer : PChar; Length : Integer) : Boolean;
   function GetExitCode : Integer;


  end;

implementation

uses debug, winioctl;

function TChildProcess.StartChild(Command : String) : Boolean;
var
   sa : TSecurityAttributes;
   si : TStartupInfo;
   hPipeOutputWrite : THandle;
   hPipeInputRead   : THandle;
begin
   Result := False;
   sa.nLength := sizeof(sa);
   sa.bInheritHandle := True;
   sa.lpSecurityDescriptor := nil;

   // Create pipe for standard output redirection.
   if not CreatePipe(hPipeOutputRead,  // read handle
              hPipeOutputWrite, // write handle
              @sa,      // security attributes
              32768) then      // number of bytes reserved for pipe - 0 default
   begin
      Log('Error creating pipe ' + IntToStr(getLastError) + SysErrorMessage(GetLastError));
   end;


   // Create pipe for standard input redirection.
   if not CreatePipe(hPipeInputRead,  // read handle
              hPipeInputWrite, // write handle
              @sa,      // security attributes
              32768) then      // number of bytes reserved for pipe - 0 default
   begin
      Log('Error creating pipe ' + IntToStr(getLastError) + SysErrorMessage(GetLastError));
   end;

   // stderr?

   // Make child process use hPipeOutputWrite as standard out,
   // and make sure it does not show on screen.
   si.cb := sizeof(si);
   si.lpReserved := nil;
   si.lpDesktop := nil;
   si.lpTitle := nil;
   si.dwFlags := STARTF_USESHOWWINDOW or STARTF_USESTDHANDLES;
   si.wShowWindow := SW_HIDE;
   si.cbReserved2 := 0;
   si.lpReserved2 := nil;
   si.hStdInput := hPipeInputRead;
   si.hStdOutput := hPipeOutputWrite;
   si.hStdError := hPipeOutputWrite;

   if CreateProcess ( nil, PChar(Command),
                   nil, nil,
                   True, 0,
                   nil, nil,
                   si, pi) then
   begin
      // Now that handles have been inherited, close then to be safe.
      // You don't want to read or write to them accidentally.
      CloseHandle(hPipeOutputWrite);
      CloseHandle(hPipeInputRead);
      Result := True;
   end
   else
   begin
      Log('Error creating process ' + IntToStr(getLastError) + SysErrorMessage(GetLastError));
      CloseHandle(hPipeOutputWrite);
      CloseHandle(hPipeOutputRead);
      CloseHandle(hPipeInputWrite);
      CloseHandle(hPipeInputRead);
   end;

end;

procedure TChildProcess.CloseChild;
begin
   // Wait for CONSPAWN to finish.
   WaitForSingleObject (pi.hProcess, INFINITE);

   // Close all remaining handles
   CloseHandle (pi.hProcess);
   CloseHandle (hPipeOutputRead);
   CloseHandle (hPipeInputWrite);
end;

function TChildProcess.GetExitCode : Integer;
var
   ExitCode : DWORD;
begin
   if GetExitCodeProcess(pi.hProcess, ExitCode) then
   begin
      Result := ExitCode;
   end
   else
   begin
      Result := -1;
   end;
end;


function TChildProcess.Read(Buffer : PChar; Length : Integer; Done : PInteger) : Boolean;
var
   Actual : DWORD;
begin
   if ReadFile2(hPipeOutputRead,      // handle of the read end of our pipe
               Buffer,            // address of buffer that receives data
               Length,                  // number of bytes to read
               Actual, // address of number of bytes read
               nil) then                  // non-overlapped.
   begin
      if Assigned(Done) then
      begin
         Done^ := Actual;
      end;
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

function TChildProcess.ReadFull(Buffer : PChar; Length : Integer) : Boolean;
var
   Remaining : Integer;
   Done : Integer;
begin
   Result := True;
   Remaining := Length;
   while Remaining > 0 do
   begin
      if self.Read(Buffer, Remaining, @Done) then
      begin
         Remaining := Remaining - Done;
         Buffer := Buffer + Done;
      end
      else
      begin
         Result := False;
         break;
      end;
   end;
end;

function TChildProcess.Write(Buffer : PChar; Length : Integer; Done : PInteger) : Boolean;
var
   Actual : DWORD;
begin
  if WriteFile2(hPipeInputWrite,
               Buffer,
               Length,
               Actual,
               nil) then
   begin
      if Assigned(Done) then
      begin
         Done^ := Actual;
      end;
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

end.
