unit lvm2;

interface

uses windows, classes, sysutils, dialogs, libext2fs, vv, types;

type lvm2_pv_head1 = record
	s_name                : array [0..7] of CHAR; // some kind of name
   s_unknown1            : U64;
   s_unknown2            : U64;
	s_magic               : array [0..7] of CHAR;  // LVM2 001
	s_uuid                : array [0..31] of CHAR;  // UUID
   s_unknown3            : U64; // size of device ?
   s_unknown4            : U64;
   s_unknown5            : U64;
   s_unknown6            : U64;
   s_unknown7            : U64;
   s_unknown8            : U64;
   s_unknown9            : U64;
   s_unknown10           : U64;
   padding1              : array [0..127] of CHAR; // pad up to 512 bytes
   padding2              : array [0..255] of CHAR; // pad up to 512 bytes
end;
type Plvm2_pv_head1 = ^lvm2_pv_head1;

type lvm2_pv_head2 = record
   s_unknown1            : ULONG;
	s_name                : array [0..3] of CHAR;
   s_unknown2            : U64;
   s_unknown4            : U64;
   s_offset1             : U64;
   s_unknown7            : U64;
   s_offset2             : U64;
   s_len1                : U64;
   s_unknown8            : U64;
   padding1              : array [0..63] of CHAR; // pad up to 512 bytes
   padding2              : array [0..127] of CHAR; // pad up to 512 bytes
   padding3              : array [0..255] of CHAR; // pad up to 512 bytes
end;
type Plvm2_pv_head2 = ^lvm2_pv_head2;

type TMetadataNode = class(TObject)
  public

   Parent : TMetadataNode;
   Name : String;
   Keys : TStringList;
   Children : TList;

   constructor Create(Parent : TMetadataNode);

   function GetPath : String;
   procedure AddKey(Key : String; Value : String);
   function FindChild(Name : String) : TMetadataNode;
   function GetValue(Key : String) : String;
   function GetValueInt(Key : String) : Integer;

end;

type TLVMGV = class(TObject)
  public
   constructor Create;
   destructor Destroy; override;

   function AddPV(Volume : TVirtualVolume; UID : String) : Integer;
   procedure ParseMetadata;
   procedure FindLV(VolumeList : TList);

  private
   PVlist : TStringList;
   PVMetadata : String;
   MetadataTree : TMetadataNode;

   lvm_head1 : lvm2_pv_head1;
   lvm_head2 : lvm2_pv_head2;
   
end;

function BuildLVM2Maps(VolumeList : TList; Options : TStringList) : Boolean;

implementation

uses native, winioctl, debug, vv_map;

const Facility = 'lvm';

{
   Notes:
      At offset 512 there is a signature 'LVM2 001'

      Collect debug info...
      You must be root.  Make a directory to collect the information into
      mkdir lvm_debug ; cd lvm_debug

      Run these commands:
      pvdisplay -vvv >  lvm_debug.txt
      vgdisplay -vvv >> lvm_debug.txt
      lvdisplay -vvv >> lvm_debug.txt

      Next, I need a copy of the metadata from your disk.  run pvdisplay and
      note which devices are physical volumes.

      For each volume, run this command, substituting DEVICE with the appropriate
      device name.  If you have multiple devices, run this command for each device

      dd if=/dev/DEVICE bs=1k count=64 of=lvm_debug.DEVICE

      Now, tar up the files and send them to me.
      cd ..
      tar -czf lvm_debug.tgz lvm_debug

      Use your mail client to send the file lvm_debug.tgz to me.
}
var
   VolumeGroups : TList;

constructor TLVMGV.Create;
begin

   pvlist := TStringList.Create;
end;

function TLVMGV.AddPV(Volume : TVirtualVolume; UID : String) : Integer;
var
   Offset   : Integer;
   Len      : Integer;

   Metadata : String;
begin
   // scan this pv for metadata

   // there is metadata at $200 -> $27F
   Volume.Read(512, sizeof(lvm_head1), @lvm_head1);
   Log('head 1:', Facility, LOG_DUMP);
   //HexDump(@lvm_head1, sizeof(lvm_head1));

   Log('I think the second header is at 0x' + IntToHex(lvm_head1.s_unknown8, 8), Facility);

   // and more at $800 -> $83F
   // well... head1:0x68 has an address...
   Volume.Read(lvm_head1.s_unknown8, sizeof(lvm_head2), @lvm_head2);
   Log('head 2:', Facility, LOG_DUMP);
   //HexDump(@lvm_head2, sizeof(lvm_head2));

   Log('metadata from ' + IntToStr(lvm_head2.s_offset1 + lvm_head2.s_offset2) + ' len ' + IntToStr(lvm_head2.s_len1), Facility, LOG_DUMP);

   Offset := lvm_head2.s_offset1 + lvm_head2.s_offset2;
   Len := lvm_head2.s_len1;

   if Offset mod 512 <> 0 then
   begin
     // and increase the len..
     Len := Len + (Offset mod 512);

     // move the offset backwards..
     Offset := Offset - (Offset mod 512);
   end;

   if Len mod 512 <> 0 then
   begin
     Len := Len + (512 - (Len mod 512));
   end;

   SetLength(Metadata, Len);

   Volume.Read(Offset, Len, PChar(Metadata));
   Metadata := AdjustLineBreaks(Metadata);
   Log(Metadata, Facility, LOG_DUMP);

   Result := 0;
   if PVlist.Count = 0 then
   begin
      PVlist.AddObject(UID, Volume);

      // The metadata should be the same on all pv in the vg, so we only
      // need to sore one copy
      PVMetadata := Metadata;
      ParseMetadata;
      
      Volume.SetLayer(Facility, LEVEL_ACTIVATED);

      Result := 1; // PV Accepted
   end
   else
   begin
      // we are adding a disk to an existing VG
      if Metadata = PVMetadata then
      begin
         // metadata matches, must be correct
         PVlist.AddObject(UID, Volume);
         Result := 1;
      end;
   end;
end;

destructor TLVMGV.Destroy;
begin

end;

procedure TLVMGV.FindLV(VolumeList : TList);
var
   i           : Integer;
   j           : Integer;
   N           : TMetadataNode;
   PVs         : TMetadataNode;
   PV          : TMetadataNode;
   PVID        : String;
   PVIndex     : Integer;
   PVOK        : Boolean;
   LV          : TMetadataNode;
   LVN         : TMetadataNode;
   SegmentCount : Integer;
   Segment     : TMetadataNode;
   Offset      : Int64;
   Start       : Int64;
   Len         : Int64;
   ExtentSize  : Int64;

   m           : TMapVolume;
begin
   if MetadataTree = nil then
   begin
      // can't do much with that...
      exit;
   end;

   if MetadataTree.Children.Count = 1 then
   begin
      N := MetadataTree.Children[0]; // VG Info
      Log('Extent size is ' + IntToStr(N.GetValueInt('extent_size') div 2) + 'k', Facility, LOG_DUMP);
      ExtentSize := N.GetValueInt('extent_size') * 512;

      // check the number of PV
      PVs := N.FindChild('physical_volumes');
      if Assigned(PVs) and (PVs.Children.Count > 1) then
      begin
         // make sure that there are enough PV and that they have the correct UID

         PVOK := True;
         for i := 0 to PVs.Children.Count - 1 do
         begin
            PV := PVs.Children[i];

            PVID := PV.GetValue('id');
            Log('Looking for PV with UID ' + PVID, Facility, LOG_VERBOSE);
            // remove the - from PVID
            PVID := StringReplace(PVID, '-', '', [rfReplaceAll]);
            PVIndex := PVlist.IndexOf(PVID);
            if PVIndex = -1 then
            begin
               Log('PV Not found', Facility, LOG_VERBOSE);
               PVOK := False;
            end
            else
            begin
               if PVlist.IndexOf(PV.Name) = -1 then // might exist from last time we tried
               begin
                  PVlist.AddObject(PV.Name, PVlist.Objects[PVIndex]);
               end;
               // do we need pe_start & pe_count
               Log('There are ' + IntToStr(PV.GetValueInt('pe_count')) + ' extents', Facility, LOG_DUMP);
            end;
         end;

         if PVOK then
         begin
            LV := N.FindChild('logical_volumes');
            if Assigned(LV) then
            begin
               Log('There are ' + IntToStr(LV.Children.Count) + ' logical volumes', Facility);

               for i := 0 to LV.Children.Count - 1 do
               begin
                  LVN := LV.Children[i];
                  Log('LV id = ' + LVN.GetValue('id'), Facility, LOG_LOOP);

                  SegmentCount := LVN.GetValueInt('segment_count');
                  if SegmentCount > 0 then
                  begin
                     m := TMapVolume.Create;
                     m.AddName(N.Name + '/' + LVN.Name);
                     VolumeList.Add(m);
                     for j := 1 to SegmentCount do
                     begin
                        Segment := LVN.FindChild('segment' + IntToStr(j));
                        if (Segment.GetValue('type') = 'striped') and
                           (Segment.GetValueInt('stripe_count') = 1) then
                        begin
                           // almost there...
                           // we need to verify which pv to use
                           PVID := Segment.GetValue('stripes[1]');
                           PVIndex := PVlist.IndexOf(PVID);
                           // offset is where in this lv = start_extent
                           // start = stripe[2]
                           // len   = extent_count
                           Offset := Segment.GetValueInt('start_extent');
                           Start  := Segment.GetValueInt('stripes[2]');
                           Len    := Segment.GetValueInt('extent_count');

                           Log(LVN.Name + ' segment ' + IntToStr(j) + ' ' + PVID + ' map ' + IntToStr(Offset) + ' -> ' + IntToStr(Start) + ',' + IntToStr(Len), Facility, LOG_VERBOSE);

                           m.AddMap(Offset * ExtentSize, Len * ExtentSize, TVirtualVolume(PVlist.Objects[PVIndex]), Start * ExtentSize + $30000); // Todo: Where does this magic number come from?
                        end
                        else
                        begin
                           Log('Wrong number of stripes (only 1 stripe supported)', Facility, LOG_ERROR);
                        end;
                     end;
                  end;
               end;
            end
            else
            begin
               Log('Can''t find logical_volumes', Facility, LOG_ERROR);
            end;
         end
         else
         begin
            Log('One or more Physical Volumes are missing or not yet detected', Facility, LOG_VERBOSE);
         end;
      end
      else
      begin
         Log('No physical volumes, how can that be?', Facility, LOG_ERROR);
      end;
   end
   else
   begin
      Log('Wrong number of volume groups', Facility, LOG_ERROR);
   end;
end;

procedure TLVMGV.ParseMetadata;
var
   L : String;
   N : TMetadataNode;
   TokenState : Integer;
   Token : String;
   TokenType : Integer;
   Tokens : TStringList;
   i : Integer;
   ParseState : Integer;
   Name : String;
   SetCount : Integer;

   procedure AcceptToken;
   begin
      if (Length(Token) > 0) or (TokenType = 1) then
      begin
         Tokens.Add(Token);
         Log('T ' + Token, Facility, LOG_LOOP);
         Token := '';
         TokenType := 0; // 0 = unknown, 1 = string constant
      end;
   end;
begin
   // this is a heirachical key = value type config file

   // constraints...
   // must only have one physical_volumes (this object)
   // we can have many logical_volumes but they must only have one segment
   //  and that segment must be a linear stripe
   //  do some maths and work out the location inside this device and then we can
   //  create some sudo partitions.
   // once all that is done, rescan for ext2/3
   // simple eh?


   // parser...
   // we should write a real parser but for now I'll cheat and just try and
   // get it going
   Tokens := TStringList.Create;
   L := PVMetadata;
   TokenState := 0;
   for i := 1 to Length(L) do
   begin
      if TokenState = 0 then
      begin
         if L[i] = '"' then
         begin
            TokenState := 1; // in quote
         end
         else if L[i] = '#' then
         begin
            TokenState := 2; // in comment
         end
         else if L[i] in [ '[', ']', ',' ] then
         begin
            // set tokens
            AcceptToken;
            Token := L[i];
            AcceptToken;
         end
         else if L[i] in [' ', #10, #13] then
         begin
            AcceptToken;
         end
         else
         begin
            Token := Token + L[i];
         end;
      end
      else if TokenState = 1 then
      begin
         TokenType := 1; // string constant
         if L[i] = '"' then
         begin
            TokenState := 0; // back to normal
         end
         else
         begin
            Token := Token + L[i];
         end;
      end
      else if TokenState = 2 then
      begin
         if L[i] in [#10, #13] then
         begin
            TokenState := 0; // back to normal
         end;
      end;
   end;


   N := TMetadataNode.Create(nil);
   N.Name := '';

   // 0 = nil
   // 1 = name
   // 2 = {
   // 3 = =

   SetCount := 0;
   ParseState := 0;
   for i := 0 to Tokens.Count - 1 do
   begin
      Token := Tokens[i];
      if ParseState = 0 then
      begin
         if Token = '}' then
         begin
            // end of node
            if Assigned(N) then
            begin
               N := N.Parent;
            end;
            ParseState := 0;
         end
         else
         begin
            Name := Token;
            ParseState := 1;
         end;
      end
      else if ParseState = 1 then
      begin
         // expect { or =
         if Token = '{' then
         begin
            // start a new node
            N := TMetadataNode.Create(N);
            N.Name := Name;
            ParseState := 0;
         end
         else if Token = '=' then
         begin
            // read the value
            ParseState := 3;
         end
         else
         begin
            Log('Parse Error. { or = expected but ' + Token + ' found', Facility, LOG_ERROR);
         end;
      end
      else if ParseState = 3 then
      begin
         if Token = '[' then
         begin
            SetCount := 1;
            ParseState := 4;
         end
         else
         begin
            //Debug(N.GetPath + Name + ' = ' + Token, DebugOff);
            N.AddKey(Name, Token);
            ParseState := 0;
         end;
      end
      else if ParseState = 4 then
      begin
         // this is a set....
         //Debug(N.GetPath + Name + '[' + IntToStr(SetCount) + '] = ' + Token, DebugOff);
         N.AddKey(Name + '[' + IntToStr(SetCount) + ']', Token);
         ParseState := 5;
      end
      else if ParseState = 5 then
      begin
         if Token = ',' then
         begin
            SetCount := SetCount + 1;
            ParseState := 4;
         end
         else if Token = ']' then
         begin
            ParseState := 0;
         end
         else
         begin
            Log('Parse Error. , or ] expected but ' + Token + ' found', Facility, LOG_ERROR);
         end;
      end;
   end;

   if N.Parent = nil then
   begin
      MetadataTree := N;
   end
   else
   begin
      Log('Incomplete metadata', Facility, LOG_ERROR);
   end;

end;

constructor TMetadataNode.Create(Parent : TMetadataNode);
begin
   self.Parent := Parent;
   Children := TList.Create;
   if Assigned(Parent) then
   begin
      Parent.Children.Add(self);
   end;
   Keys := TStringList.Create;
end;

function TMetadataNode.GetPath : String;
begin
   Result := '';
   if Assigned(Parent) then
   begin
      Result := Parent.GetPath;
   end;
   Result := Result + Name + '\';
end;

procedure TMetadataNode.AddKey(Key : String; Value : String);
begin
   Keys.Add(Key + '=' + Value);
end;

function TMetadataNode.GetValue(Key : String) : String;
begin
   Result := Keys.Values[key];
end;

function TMetadataNode.GetValueInt(Key : String) : Integer;
begin
   try
      Result := StrToInt(Trim(Keys.Values[key]));
   except
      on E : Exception do
      begin
         ShowMessage(E.Message + #10 + 'While looking up ' + Name + '\' + Key);
         raise E;
      end;
   end;
end;

function TMetadataNode.FindChild(Name : String) : TMetadataNode;
var
   i : Integer;
begin
   Result := nil;
   for i := 0 to Children.Count - 1 do
   begin
      if TMetadataNode(Children[i]).Name = Name then
      begin
         Result := TMetadataNode(Children[i]);
         break;
      end
   end;
end;

function BuildLVM2Maps(VolumeList : TList; Options : TStringList) : Boolean;
var
   i : Integer;
   j : Integer;
   v : TVirtualVolume;

   Buffer      : String;
   Magic       : String;
   UID         : String;

   VG          : TLVMGV;
   Added       : Boolean;

begin
   Result := False;

   if VolumeGroups = nil then
   begin
      VolumeGroups := TList.Create;
   end;

   for i := 0 to VolumeList.Count - 1 do
   begin
      v := TVirtualVolume(VolumeList[i]);
      Log('Scanning ' + v.GetName, 'lvm', LOG_VERBOSE);

      if v.GetLayer('lvm2') = LEVEL_NONE then
      begin
         v.SetLayer('lvm2', LEVEL_QUERIED);
         SetLength(Buffer, 512);
         if v.Read(512, 512, PChar(Buffer)) then
         begin
            Magic := Copy(Buffer, 25, 8);
            if Magic = 'LVM2 001' then
            begin
               Log('LVM2 Found. magic = ' + Magic, 'lvm', LOG_VERBOSE);
               v.SetLayer('lvm2', LEVEL_PROCESSED);

               UID := Copy(Buffer, 33, 32);
               Log('PV UID = ' + UID, 'lvm', LOG_VERBOSE);
               v.AddName('LVM2 PV ' + UID);

               // is there a volume group waiting for this PV?
               Added := False;
               for j := 0 to VolumeGroups.Count - 1 do
               begin
                  VG := VolumeGroups[j];
                  if VG.AddPV(v, UID) > 0 then
                  begin
                     Log('PV Added to existing VG', Facility, LOG_VERBOSE);
                     Added := True;
                     VG.FindLV(VolumeList);
                     break;
                  end;
               end;

               if not Added then
               begin
                  // create a new VG for this PV
                  VG := TLVMGV.Create;
                  VolumeGroups.Add(VG);
                  VG.AddPV(v, UID);
                  VG.FindLV(VolumeList);
               end;
               Result := True;
            end
         end;
      end;
   end;
end;


initialization
   VolumeGroups := nil;
   RegisterVolumeScan(Facility, BuildLVM2Maps);

end.
