unit vfs;

interface

uses
types, vv, windows, classes;

var
   NodeCount : Integer;
type
   TMountOptions = class(TObject)
   private
      Data : TStringList;
   public
      constructor Create;
      destructor Destroy; override;

      procedure Add(Option : String);

      function GetParameter(Index : Integer) : String;
      function HasOption(Name : String) : Boolean;
      function GetOption(Name : String) : String;
      function GetFlag(Name : String) : Boolean;

   end;

   TVNode = class;

   TFs = class(TObject)
   public
      constructor Create; virtual;
      destructor Destroy; override;

      function IsWritable : Boolean; virtual;

      function Mount(Options : TMountOptions) : TVNode; virtual; abstract;
      function UMount : Boolean; virtual; abstract;

   end;

   TFsClass = class of TFs;

   // A Filesystem which has a block based store (ie a disk fs)
   TVVFs = class(TFs)
   public
      v : TVirtualVolume;

      procedure AssignVolume(Volume : TVirtualVolume);
   end;

   TVNode = class(TObject)
   public
      fs : TFs;
      Name  : String;

      Parent      : TVNode;
      Overlay     : TVNode;
      MountPoint  : TVNode;
      UMounting   : Boolean;

      // test mounting stuff

      constructor Create; virtual;
      destructor Destroy; override;

      function HasChildren : Boolean; virtual; abstract; // is diectory ?
      function HasData : Boolean; virtual; abstract; // is file?
      function GetLen : Int64; virtual; abstract;
      function GetAttribute(Attribute : String) : String; virtual;

      // Read/Write functions
      // SetLen
      function Mkdir(Name : String) : Boolean; virtual; abstract;
      function CreateFile(Name : String) : Boolean; virtual; abstract;
      function Delete(Name : String) : Boolean; virtual; abstract;

      function OpenData(WriteAccess : Boolean) : Boolean; virtual; abstract;
      function IsWritable : Boolean; virtual;
      function CloseData : Boolean; virtual; abstract;
      function ReadData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean; virtual; abstract;
      function WriteData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean; virtual; abstract;
      function SetLen(Len : Int64) : Boolean; virtual; abstract;

      // Testing procedures
      procedure DumpDir(Prefix : String; Recurse : Boolean);

      // Does this belong here?
      procedure ListChildren(Children : TStringList); virtual; abstract;
      function HasParent : Boolean;
      function GetParentNode : TVNode;
      function GetChildNode(Name : String) : TVNode;
      procedure AddChildNode(Node : TVNode);

      function FindNode(Path : String) : TVNode;

      function Mount(fs : TFs; Options : TMountOptions) : Boolean;
      function UMount : Boolean;
      procedure Close;
      function Open : TVNode;
   protected
      Cache    : TStringList;
      RefCount : Integer;

      function GetChild(Name : String) : TVNode; virtual; abstract;
      procedure RemoveCache(Node : TVNode);
   end;


var
   FileSystems : TStringList;

procedure RegisterFileSystem(Name : String; ClassName : TFsClass);


implementation

uses sysutils, vv_io, debug, unicode;

const Facility = 'vfs';

constructor TFs.Create;
begin
end;

destructor TFs.Destroy;
begin
end;

constructor TVNode.Create;
begin
   Cache := nil;
   RefCount := 0;
   Inc(NodeCount);
end;

destructor TVNode.Destroy;
begin
   Dec(NodeCount);
   if Assigned(Parent) then
   begin
      Parent.RemoveCache(self);
      Parent.Close;
      Parent := nil;
   end;
   if Assigned(Cache) then
   begin
      Cache.Free;
      Cache := nil;
   end;
end;

function TVNode.Open : TVNode;
begin
   if RefCount = 0 then
   begin
      //Log('Opening a deleted VNode.  This can''t happen');
      // Unless I have just been created...
   end;
   if Assigned(Overlay) then
   begin
      //Log('Opening a mount point.  This can''t happen');
      // Unless I have just been created...
   end;

   Inc(RefCount);
   Result := self;
end;

procedure TVNode.Close;
begin
   Dec(RefCount);
   if UMounting and (RefCount = 1) then
   begin
      Dec(RefCount);
      Log('umounting filesysem', Facility, LOG_VERBOSE);
      fs.UMount;
      fs.Free;
   end;
   if RefCount = 0 then
   begin
//      Log('RefCount is ' + IntToStr(RefCount) + ' for ' + Name);
      Free;
   end;
end;

function TVNode.Mount(fs : TFs; Options : TMountOptions) : Boolean;
var
   i : Integer;
begin
   Result := False;

   Overlay := fs.Mount(Options);
   if Assigned(Overlay) then
   begin
      Overlay.Open;
      Overlay.MountPoint := self.Open;
      if Assigned(Parent) then
      begin
         i := Parent.Cache.IndexOf(Name);
         if i >=0 then
         begin
            Parent.Cache.Objects[i] := Overlay;
            Overlay.Name := Name;
            Result := True;
         end
         else
         begin
            Log('Error! I must exist', Facility, LOG_ERROR);
         end
      end
      else
      begin
         // mount as new root
         Result := True;
      end;
   end;
end;

function TVNode.UMount : Boolean;
var
   i : Integer;
begin
   Result := False;

   if Assigned(MountPoint) then
   begin
      if MountPoint.Overlay = self then
      begin
         i := MountPoint.Parent.Cache.IndexOf(Name);
         if i >=0 then
         begin
            MountPoint.Parent.Cache.Objects[i] := MountPoint;
         end
         else
         begin
            Log('Error! I must exist', Facility, LOG_ERROR);
         end;
         Parent := MountPoint.Parent; // link this here because at 'cd' time the overlay pointer will be gone
         MountPoint.Overlay := nil;
         MountPoint.Close;
         MountPoint := nil;
         // we need to fs.UMount when the ref count is decremented
         UMounting := True;
         self.Close;
      end;
   end;
end;

function TVNode.HasParent : Boolean;
begin
   Result := False;
   if Assigned(Parent) then
   begin
      Result := True;
   end
   else
   begin
      if Assigned(MountPoint) then
      begin
         if Assigned(MountPoint.Parent) then
         begin
            Result := True;
         end;
      end;
   end;
end;

// the result of this function must be 'close'd when done
function TVNode.GetParentNode : TVNode;
begin
   if Assigned(Parent) then
   begin
      Result := Parent.Open;
   end
   else if Assigned(MountPoint) then
   begin
      if Assigned(MountPoint.Parent) then
      begin
         Result := MountPoint.Parent.Open;
      end
      else
      begin
         Result := self.Open;
      end;
   end
   else
   begin
      Result := self.Open;
   end;
end;

// This will look for an existing node object
// and if there is not one, create it with GetChild
// the result of this function must be 'close'd when done
function TVNode.GetChildNode(Name : String) : TVNode;
var
   i : Integer;
begin
   if not Assigned(Cache) then
   begin
      Cache := TStringList.Create;
      Cache.Sorted := True;
   end;

   i := Cache.IndexOf(Name);
   if i >=0 then
   begin
      Result := TVNode(Cache.Objects[i]).Open;
   end
   else
   begin
      if Assigned(Overlay) then
      begin
         Result := Overlay.GetChild(Name);
      end
      else
      begin
         Result := GetChild(Name);
      end;
      if Assigned(Result) then
      begin
         Result.Open;
         Result.Parent := self.Open;
         Cache.AddObject(Name, Result);
      end;
   end;
end;

procedure TVNode.AddChildNode(Node : TVNode);
var
   i : Integer;
begin
   if not Assigned(Cache) then
   begin
      Cache := TStringList.Create;
      Cache.Sorted := True;
   end;
   i := Cache.IndexOf(Node.Name);
   if i >=0 then
   begin
      // we already have an entry.. use the old or the new?
   end
   else
   begin
      Node.Open;
      Node.Parent := self.Open;
      Cache.AddObject(Node.Name, Node);
   end;
end;


procedure TVNode.RemoveCache(Node : TVNode);
var
   i : Integer;
begin
   i := Cache.IndexOf(Node.Name);
   if i >=0 then
   begin
      Cache.Delete(i);
   end
   else
   begin
      Log('It''s not here!', Facility, LOG_ERROR);
   end;
end;

{
   Takes a relative or absolute path.
   relative paths are looked up component by component
   absolute paths have to start at the vfs root.
}
function TVNode.FindNode(Path : String) : TVNode;
var
   p : Integer;
   Name : String;
   Next : TVNode;
begin
   Result := self.Open;

   while Length(Path) > 0 do
   begin
      p := Pos('/', Path);
      if p = 0 then // last part
      begin
         Name := Path;
         Path := '';
      end
      else if p = 1 then // absolute path
      begin
         Name := '/';
         Path := Copy(Path, 2, Length(Path));
      end
      else
      begin
         Name := Copy(Path, 1, p - 1);
         Path := Copy(Path, p + 1, Length(Path));
      end;

      // Process Name
      Log('- ' + Name, Facility, LOG_VERBOSE);

      if Name = '.' then
      begin
         // Nothing to do
      end
      else if Name = '..' then
      begin
         Next := Result.GetParentNode;
         Result.Close;
         Result := Next;
      end
      else if Name = '/' then
      begin
         while Result.HasParent do
         begin
            Next := Result.GetParentNode;
            Result.Close;
            Result := Next;
         end;
      end
      else
      begin
         Next := Result.GetChildNode(Name);
         Result.Close;
         Result := Next;
      end;

      if not Assigned(Result) then
      begin
         Log('File not found (' + Name + ')', Facility, LOG_NORMAL);
         break;
      end;

   end;
end;

procedure TVNode.DumpDir(Prefix : String; Recurse : Boolean);
var
   Node : TVNode;
   Children : TStringList;
   i : Integer;
begin
{   if HasData then
   begin
      LogW(Encode(Name) + ' size is ' + IntToStr(GetLen));
   end
   else if HasChildren then
   begin
      LogW('Dir ' + Encode(Name));
      Children := TStringList.Create;
      ListChildren(Children);
      for i := 0 to Children.Count - 1 do
      begin
         if Children[i] = '.' then
         begin
            // do nothing
         end
         else if Children[i] = '..' then
         begin
            // do nothing
         end
         else
         begin
            Node := GetChild(Children[i]);
            if Assigned(Node) then
            begin
               Node.DumpDir;
               Node.Free;
            end;
         end;
      end;
      Children.Free;
   end;}
   Prefix := Prefix + '[' + Name + '] ';
   Log(Prefix + ' ref count = ' + IntToStr(RefCount));
   if not Assigned(Parent) then
   begin
      Log(Prefix + 'I am the root of my FS');
   end;
   if Assigned(Overlay) then
   begin
      Log(Prefix + 'Something is mounted here');
   end;
   if Assigned(MountPoint) then
   begin
      Log(Prefix + 'I am mounted on someone');
   end;
   if Assigned(Cache) then
   begin
      if Cache.Count > 0 then
      begin
         for i := 0 to Cache.Count - 1 do
         begin
            Log(Prefix + 'Cache entry ' + Cache[i]);
            if Recurse then
            begin
               with Cache.Objects[i] as TVNode do begin
                  DumpDir(Prefix, Recurse);
               end;
            end;
         end;
      end
      else
      begin
         Log(Prefix + 'Cache is empty');
      end;
   end;
end;

function TVNode.IsWritable : Boolean;
begin
   Result := fs.IsWritable;
end;

function TVNode.GetAttribute(Attribute : String) : String;
begin
   Result := '';
end;


function TFs.IsWritable : Boolean;
begin
   Result := False;
end;

procedure TVVFs.AssignVolume(Volume : TVirtualVolume);
begin
   v := Volume;
end;

constructor TMountOptions.Create;
begin
   inherited;

   Data := TStringList.Create;
end;

destructor TMountOptions.Destroy;
begin
   Data.Free;

   inherited;
end;

procedure TMountOptions.Add(Option : String);
begin
   Data.Add(Option);
end;

function TMountOptions.HasOption(Name : String) : Boolean;
begin
   if Data.IndexOfName(Name) >= 0 then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

function TMountOptions.GetOption(Name : String) : String;
begin
   Result := Data.Values[Name];
end;

function TMountOptions.GetParameter(Index : Integer) : String;
begin
   if Data.Count > Index then
   begin
      Result := Data[Index];
   end
   else
   begin
      Result := '';
   end
end;

function TMountOptions.GetFlag(Name : String) : Boolean;
begin
   if Data.IndexOf(Name) >= 0 then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;



// For dynamic fs type registration
procedure RegisterFileSystem(Name : String; ClassName : TFsClass);
begin
   if not Assigned(FileSystems) then
   begin
      FileSystems := TStringList.Create;
   end;
   FileSystems.AddObject(Name, TObject(ClassName));
end;


initialization
NodeCount := 0;

end.
