unit vfs_ext2;

interface
uses types, vfs, libext2fs, classes, vv_io;

type
   TExtFs = class(TVVFs)
   private
      WriteAccess : Boolean;
   public
      fs       : Pext2_filsys;
      function Mount(Options : TMountOptions) : TVNode; override;
      function UMount : Boolean; override;

   end;

   TExtNode = class(TVNode)
   public
      constructor Create; override;
      destructor Destroy; override;

      function HasChildren : Boolean; override;
      function HasData : Boolean; override;
      function GetLen : U64; override;
      function GetAttribute(Attribute : String) : String; override;


      procedure ListChildren(Children : TStringList); override;
      function Mkdir(Name : String) : Boolean; override;
      function CreateFile(Name : String) : Boolean; override;
      function Delete(Name : String) : Boolean; override;


      function OpenData(WriteAccess : Boolean) : Boolean; override;
      function CloseData : Boolean; override;
      function ReadData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean; override;
      function WriteData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean; override;
      function SetLen(Len : Int64) : Boolean; override;

   protected
      ino   : ext2_ino_t;
      INode : ext2_inode;

      fp : Pext2_file_t;
      fp_dirty : Boolean;

      function GetChild(Name : String) : TVNode; override;

      function ReadInode : errcode_t;
      function WriteInode : errcode_t;
      procedure Flush;
   end;

implementation

uses windows, sysutils, debug, unicode, time, vv, ls;

const Facility = 'vfs_ext2';

type TFindEntry = record
   Name : PChar;
   Len : Integer;
   ino : ext2_ino_t;
end;
PFindEntry = ^TFindEntry;

function find_dir_entry(dirent : Pext2_dir_entry; offset : Integer; blocksize : Integer; buf : PChar; priv_data : Pointer) : Integer; cdecl;
var
   Find  : PFindEntry;
begin
   Result := 0;
   Find := PFindEntry(priv_data);

   if dirent.name_len = Find.Len then
   begin
      if CompareMem(@dirent.name[0], Find.Name, dirent.name_len) then
      begin
         Find.ino := dirent.inode;
         Result := DIRENT_ABORT;
      end;
   end;
end;

function on_dir_entry(dirent : Pext2_dir_entry; offset : Integer; blocksize : Integer; buf : PChar; priv_data : Pointer) : Integer; cdecl;
var
   Name     : String;
   Children : TStringList;
begin
   SetLength(Name, dirent.name_len);
   CopyMemory(PChar(Name), @dirent.name[0], dirent.name_len);
   Children := TStringList(priv_data);
   if Assigned(Children) then
   begin
      Children.AddObject(Name, Pointer(dirent.inode));
   end;
   Result := 0;
end;


constructor TExtNode.Create;
begin
   inherited;

   fp := nil;
end;

destructor TExtNode.Destroy;
begin
   CloseData;

   inherited;
end;


function TExtNode.HasChildren : Boolean;
begin
   if (INode.i_mode and S_IFMT) = S_IFDIR then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

function TExtNode.HasData : Boolean;
begin
   if (INode.i_mode and S_IFMT) = S_IFREG then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

function TExtNode.GetLen : U64;
var
   Size : LARGE_INTEGER;
begin
   Size.LowPart := INode.i_size;
   Size.HighPart := INode.i_dir_acl;
   Result := Size.QuadPart;
end;


function TExtNode.GetAttribute(Attribute : String) : String;
var
   Temp : LARGE_INTEGER;
begin
   if Attribute = 'system' then
   begin
      Result := 'ext2';
   end
   else if Attribute = 'user' then
   begin
      Temp.LowPart := INode.i_uid;
      Temp.HighPart := INode.l_i_uid_high;
      Result := IntToStr(Temp.QuadPart);
   end
   else if Attribute = 'group' then
   begin
      Temp.LowPart := INode.i_gid;
      Temp.HighPart := INode.l_i_gid_high;
      Result := IntToStr(Temp.QuadPart);
   end
   else if Attribute = 'ctime' then
   begin
      Result := CTime(INode.i_ctime);
   end
   else if Attribute = 'mtime' then
   begin
      Result := CTime(INode.i_mtime);
   end
   else if Attribute = 'atime' then
   begin
      Result := CTime(INode.i_atime);
   end
   else if Attribute = 'mode' then
   begin
      Result := GetModeString(INode.i_mode);
   end
end;


function TExtNode.GetChild(Name : String) : TVNode;
var
   e     : errcode_t;
   Find  : TFindEntry;
   Node  : TExtNode;
begin
   Result := nil;

   Find.Name   := PChar(Name);
   Find.Len    := Length(Name);
   Find.ino    := 0;
   ext2fs_dir_iterate(TExtFs(fs).fs, ino, 0, nil, find_dir_entry, @Find);
   if Find.ino > 0 then
   begin
      Node := TExtNode.Create;
      Node.ino    := Find.ino;
      Node.Name   := Name;
      Node.fs     := fs;
      e := Node.ReadInode;
      if e = 0 then
      begin
         Result := Node;
      end
      else
      begin
         Log('Error ' + IntToStr(e));
         Node.Free;
      end;
   end;
end;

procedure TExtNode.ListChildren(Children : TStringList);
var
   extfs : TExtFs;
begin
   extfs := TExtFs(fs);
   ext2fs_dir_iterate(extfs.fs, ino, 0, nil, on_dir_entry, Children);
end;

function TExtNode.ReadInode : errcode_t;
begin
   Result := ext2fs_read_inode(TExtFs(fs).fs, ino, @INode);
end;

function TExtNode.WriteInode : errcode_t;
begin
   Result := ext2fs_write_inode(TExtFs(fs).fs, ino, @INode);
end;

function TExtNode.Mkdir(Name : String) : Boolean;
var
   Node        : TVNode;
   e           : errcode_t;
   new_inode   : ext2_ino_t;
begin
   Result := False;
   Node := GetChildNode(Name);
   if Assigned(Node) then
   begin
      Log('File already exists');
      Node.Close;
   end
   else
   begin
      new_inode := 0;
      e := ext2fs_new_inode(TExtFs(fs).fs, Self.ino, S_IFDIR or S_IRUSR or S_IWUSR or S_IXUSR or S_IRGRP or S_IXGRP or S_IROTH or S_IXOTH, nil, @new_inode);
      if e = 0 then
      begin
         Log('New dir inode will be ' + IntToStr(new_inode), Facility, LOG_VERBOSE);
         if new_inode = ino then
         begin
            Log('This is my inode!', Facility, LOG_ERROR);
            exit;
         end;
         Log('making directory ' + Name + ' in dir inode ' + IntToStr(ino), Facility, LOG_VERBOSE);
         e := ext2fs_mkdir(TExtFs(fs).fs, ino, new_inode, PChar(Name));
         if e = EXT2_ET_DIR_NO_SPACE then
         begin
            e := ext2fs_expand_dir(TExtFs(fs).fs, ino);
            if e = 0 then
            begin
               e := ext2fs_mkdir(TExtFs(fs).fs, ino, 0, PChar(Name));
            end;
         end;
         if e = 0 then
         begin
            Result := True;
            Flush;
         end
         else
         begin
            Log('error ' + intToStr(e));
         end;
      end;
   end;
end;

function TExtNode.CreateFile(Name : String) : Boolean;
var
   Node        : TVNode;
   e           : errcode_t;
   new_inode   : ext2_ino_t;
   inode_data  : ext2_inode;
begin
   Result := False;

   Node := GetChildNode(Name);
   if Assigned(Node) then
   begin
      Log('File already exists');
      Node.Close;
   end
   else
   begin
      new_inode := 0;
      e := ext2fs_new_inode(TExtFs(fs).fs, Self.ino, S_IFREG or S_IRUSR or S_IWUSR or S_IXUSR or S_IRGRP or S_IXGRP or S_IROTH or S_IXOTH, nil, @new_inode);
      if e = 0 then
      begin
         Log('New inode is ' + IntToStr(new_inode), Facility, LOG_VERBOSE);
         if new_inode = ino then
         begin
            Log('This is my inode!', Facility, LOG_ERROR);
            exit;
         end;
         // link into parent directory (self)
         e := ext2fs_link(TExtFs(fs).fs, ino, PChar(Name), new_inode, EXT2_FT_REG_FILE);
         if e = EXT2_ET_DIR_NO_SPACE then
         begin
            e := ext2fs_expand_dir(TExtFs(fs).fs, ino);
            if e = 0 then
            begin
               e := ext2fs_link(TExtFs(fs).fs, ino, PChar(Name), new_inode, EXT2_FT_REG_FILE);
            end;
         end;

         if e = 0 then
         begin
            // we are linked in, save the inode
            FillMemory(@inode_data, sizeof(inode_data), 0);
            inode_data.i_mode := S_IFREG or S_IRUSR or S_IWUSR or S_IRGRP or S_IROTH;
            inode_data.i_links_count := 1;
            inode_data.i_atime := CurrentCTime;
            inode_data.i_ctime := inode_data.i_atime;
            inode_data.i_mtime := inode_data.i_atime;
            e := ext2fs_write_new_inode(TExtFs(fs).fs, new_inode, @inode_data);
            if e = 0 then
            begin
               ext2fs_inode_alloc_stats(TExtFs(fs).fs, new_inode, 1);
               Flush;
               Result := True;
            end
            else
            begin
               // we need to unlink...
            end;
         end;
      end
      else
      begin
         Log('Error opening file ' + IntToStr(e));
         fp := nil;
      end;
   end;
end;

function on_block_release(fs : Pext2_filsys; blocknr : Pblk_t; blockcnt : Integer; private : Pointer) : Integer; cdecl;
begin
   Log('Processing block ' + IntToStr(blocknr^), Facility, LOG_LOOP);
	ext2fs_block_alloc_stats(fs, blocknr^, -1);
   Result := 0;
end;

function TExtNode.Delete(Name : String) : Boolean;
var
   Node     : TVNode;
   ExtNode  : TExtNode;
   e        : errcode_t;
begin
   Result := False;
   // Make sure the file exists.
   Node := GetChildNode(Name);
   if Assigned(Node) then
   begin
      if Node is TExtNode then
      begin
         if Node.HasData then
         begin
            ExtNode := TExtNode(Node);
            ExtNode.INode.i_links_count := ExtNode.INode.i_links_count - 1;
            e := ExtNode.WriteInode;
            if e = 0 then
            begin
               Result := True; // if it goes wrong now, we get corruption
               // we need name because the one inode might have 2 hard links in
               // the one directory.  If we use name, we dont need the inode
               // number any more, but we throw it in for good measure.
               e := ext2fs_unlink(TExtFs(fs).fs, ino, PChar(name), ExtNode.ino, 0); // we need to make sure we are on the same device...
               if e <> 0 then
               begin
                  Log('Error unlinking inode');
               end;
               if ExtNode.INode.i_links_count = 0 then
               begin
                  ExtNode.INode.i_dtime := CurrentCTime;
                  e := ExtNode.WriteInode;
                  if e <> 0 then
                  begin
                     Log('Error deleting inode');
                  end;

                  // we need to free the blocks
                  ext2fs_block_iterate(TExtFs(fs).fs, ExtNode.ino, 0, nil, @on_block_release, nil);
                  ext2fs_inode_alloc_stats(TExtFs(fs).fs, ExtNode.ino, -1);
               end;
               Flush;
            end;
         end
         else
         begin
            Log('Deleting directories is NYI');
         end;
      end;
      Node.Close;
   end;
end;

function TExtNode.OpenData(WriteAccess : Boolean) : Boolean;
var
   e        : errcode_t;
   Flags    : Integer;
begin
   Result := False;

   if Assigned(fp) then
   begin
      // promote read to write?
      Result := True;
   end
   else
   begin
      Flags := 0;
      if WriteAccess then
      begin
         Flags := Flags or EXT2_FILE_WRITE;
      end;
      e := ext2fs_file_open(TExtFs(fs).fs, ino, Flags, @fp);
      if e = 0 then
      begin
         Result := True;
         fp_dirty := False;
      end
      else
      begin
         Log('Error opening file ' + IntToStr(e));
         fp := nil;
      end;
   end;
end;

function TExtNode.CloseData : Boolean;
var
   e        : errcode_t;
begin
   if Assigned(fp) then
   begin
      // if we are writing then we need to maintain the inode ourself

      if fp_dirty then
      begin
         e := ext2fs_write_inode(TExtFs(fs).fs, fp.ino, @fp.inode);
         if e = 0 then
         begin
         end
         else
         begin
            Log('Error saving inode');
         end;
      end;
      ext2fs_file_close(fp);
      fp := nil;
      Flush;
   end;
   Result := True;
end;

function TExtNode.ReadData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean;
var
   e        : errcode_t;
   NewPos   : U64;
   Got      : U32;
begin
   Result := False;

   OpenData(False);

   if Assigned(fp) then
   begin
      e := ext2fs_file_llseek(fp, Offset, EXT2_SEEK_SET, @NewPos);
      if e <> 0 then
      begin
         Log('Error seeking in file ' + IntToStr(e));
         exit;
      end;
      e := ext2fs_file_read(fp, Buffer, Len, @Got);
      if e <> 0 then
      begin
         Log('Error reading file ' + IntToStr(e));
      end
      else
      begin
         if Got = Len then
         begin
            Result := True;
         end
         else
         begin
            Log('Only read ' + IntToStr(Got) + ' bytes');
         end;
      end;
   end;
end;

function TExtNode.WriteData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean;
var
   e        : errcode_t;
   NewPos   : U64;
   Got      : U32;
   isize    : LARGE_INTEGER;
begin
   Result := False;

   OpenData(True);

   if Assigned(fp) then
   begin
      e := ext2fs_file_llseek(fp, Offset, EXT2_SEEK_SET, @NewPos);
      if e <> 0 then
      begin
         Log('Error seeking in file ' + IntToStr(e));
         exit;
      end;
      e := ext2fs_file_write(fp, Buffer, Len, @Got);
      if e <> 0 then
      begin
         Log('Error writing file ' + IntToStr(e));
      end
      else
      begin
         if Got = Len then
         begin
            Result := True;

            isize.LowPart := fp.inode.i_size;
            isize.HighPart := fp.inode.i_dir_acl;
            if Offset + Len > isize.QuadPart then
            begin
               // we need to extend the file size
               isize.QuadPart     := Offset + Len;
               fp.inode.i_size    := isize.LowPart;
               fp.inode.i_dir_acl := isize.HighPart;
               fp_dirty := True;
            end;
         end
         else
         begin
            Log('Only wrote ' + IntToStr(Got) + ' bytes');
         end;
      end;
   end;
end;

function TExtNode.SetLen(Len : Int64) : Boolean;
begin
   Result := False;
end;

procedure TExtNode.Flush;
var
   e     : errcode_t;
begin
   if TExtFs(fs).WriteAccess then
   begin
      if TExtFs(fs).fs.super.s_state = 1 then
      begin
         TExtFs(fs).fs.super.s_state := 0;// what value?  Is this a bit mask or a state?
      end;
      e := ext2fs_write_bitmaps(TExtFs(fs).fs);
      if e <> 0 then
      begin
         Log('Error writing bitmaps ' + IntToStr(e), Facility, LOG_ERROR);
      end;
      e := ext2fs_flush(TExtFs(fs).fs);
      if e <> 0 then
      begin
         Log('Error flushing filesystem ' + IntToStr(e), Facility, LOG_ERROR);
      end;
   end;
end;

function TExtFs.Mount(Options : TMountOptions) : TVNode;
var
   e     : errcode_t;
   Flags : Integer;
begin
   Result := nil;
   Flags := 0;
   if Options.GetFlag('rw') then
   begin
      Flags := Flags or EXT2_FLAG_RW;
   end;
   e := ext2fs_open(PChar(v.GetUniqueName), Flags, 0, 0, get_vv_io_manager, @fs);
   if e = 0 then
   begin
      Result := TExtNode.Create;
      with Result as TExtNode do
      begin
         ino    := EXT2_ROOT_INO;
         Name   := '/';
         fs     := Self;
         e      := ReadInode;
      end;
      if e = 0 then
      begin
         Result.Open;

         if Options.GetFlag('rw') then
         begin
            e := ext2fs_read_bitmaps(fs);
            if e = 0 then
            begin
               WriteAccess := True;
            end
            else
            begin
               Log('ext2fs_read_bitmaps error ' + intToStr(e), Facility, LOG_ERROR);
            end;
         end;
      end
      else
      begin
         Log('Error ' + IntToStr(e));
         Result.Free;
         Result := nil;
      end;
   end
   else
   begin
      Log('Error ' + IntToStr(e));
   end;
end;

function TExtFs.UMount : Boolean;
var
   e : errcode_t;
begin
   Result := False;
   //Flush;
   e := ext2fs_close(fs);
   if e = 0 then
   begin
      Result := True;
   end
   else
   begin
      Log('Error ' + IntToStr(e));
   end;
end;

// http://www.solarisinternals.com/si/reading/vnode.pdf

// if we find any ext2, add an alias with the volume name
function BuildExt2Maps(VolumeList : TList; Options : TStringList) : Boolean;
var
   i : Integer;
   v : TVirtualVolume;

   Superblock : ext2_super_block;
   UID : String;
   Name : String;
begin
   Result := False;

   for i := 0 to VolumeList.Count - 1 do
   begin
      v := TVirtualVolume(VolumeList[i]);

      if v.GetLayer(Facility) = LEVEL_NONE then
      begin

         // we don't want to process RAID member disks...
         if v.GetLayer('raid') = LEVEL_QUERIED then
         begin
            v.SetLayer(Facility, LEVEL_QUERIED);
            if v.Read(1024, sizeof(Superblock), @Superblock) then
            begin
               if Superblock.s_magic = EXT2_SUPER_MAGIC then
               begin
                  Log('Ext2 Found. magic = 0x' + IntToHex(Superblock.s_magic, 4), Facility, LOG_VERBOSE);
                  v.SetLayer('lvm2', LEVEL_PROCESSED);
                  v.AddName('FS=ext2');

                  if Superblock.s_rev_level >= EXT2_DYNAMIC_REV then
                  begin
                     UID := UuidToStr(Superblock.s_uuid);
                     v.AddName('UUID=' + UID);
                     Name := Trim(Superblock.s_volume_name);
                     if Length(Name) > 0 then
                     begin
                        v.AddName('LABEL=' + Name);
                     end;
                  end;
               end
            end;
         end
         else if v.GetLayer('raid') = LEVEL_NONE then
         begin
            Log('Not scanning unchecked raid member ' + v.GetName, Facility, LOG_LOOP);
            Result := True;
         end
         else if v.GetLayer('raid') = LEVEL_PROCESSED then
         begin
            Log('Not scanning unused raid member ' + v.GetName, Facility, LOG_LOOP);
         end
         else if v.GetLayer('raid') = LEVEL_ACTIVATED then
         begin
            Log('Not scanning used raid member ' + v.GetName, Facility, LOG_LOOP);
         end
         else
         begin
            Log('Not scanning unknown status raid ' + v.GetName + ' at level ' + IntToStr(v.GetLayer('raid')), Facility, LOG_VERBOSE);
         end;
      end;
   end;
end;

initialization
   RegisterVolumeScan(Facility, BuildExt2Maps);
   RegisterFileSystem('ext2', TExtFs);
   RegisterFileSystem('ext3', TExtFs);


end.
