unit vfs_fat;

interface

uses types, vfs, libfat, classes, vv_io;

type

   TFatFs = class(TVVFs)
   private
      WriteAccess : Boolean;
   public
      fs       : PFatFS;
      function Mount(Options : TMountOptions) : TVNode; override;
      function UMount : Boolean; override;

   end;

   TFatNode = class(TVNode)
   public
      constructor Create; override;
      destructor Destroy; override;

      function HasChildren : Boolean; override;
      function HasData : Boolean; override;
      function GetLen : U64; override;
      function GetAttribute(Attribute : String) : String; override;


      procedure ListChildren(Children : TStringList); override;
      {function Mkdir(Name : String) : Boolean; override;
      function CreateFile(Name : String) : Boolean; override;
      function Delete(Name : String) : Boolean; override;}


      function OpenData(WriteAccess : Boolean) : Boolean; override;
      function CloseData : Boolean; override;
      function ReadData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean; override;
{      function WriteData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean; override;
      function SetLen(Len : Int64) : Boolean; override;}

   protected
      h : TFatFile;
      Cluster  : Integer;
      Directory : Integer; // cluster that we are listed in
      ShortName : String;
      Size     : U32;
      Attrib   : Integer;

      function GetChild(Name : String) : TVNode; override;

   end;

implementation

uses windows, sysutils, debug, unicode, time, vv, ls;

const Facility = 'vfs_fat';

type TFindEntry = record
   Name : String;
   attr : TFatAttr;
end;
PFindEntry = ^TFindEntry;

function find_dir_entry(attr : PFatAttr; priv_data : Pointer) : Integer; cdecl;
var
   Find  : PFindEntry;
   Match : Boolean;
begin
   Result := 0;
   Match := False;
   Find := PFindEntry(priv_data);

   if CompareText(attr.ShortName, Find.Name) = 0 then
   begin
      Match := True;
   end
   else if Assigned(attr.LongName) then
   begin
      if attr.LongName = Find.Name then
      begin
         Match := True;
      end
   end;

   if Match then
   begin
      CopyMemory(@Find.attr, attr, sizeof(TFatAttr));
      Result := 2; // found it
   end;

end;

function on_dir_entry(attr : PFatAttr; priv_data : Pointer) : Integer; cdecl;
var
   Children : TStringList;
begin
   Children := TStringList(priv_data);
   if Assigned(Children) then
   begin
      if Assigned(attr.LongName) then
      begin
         Children.Add(attr.LongName);
      end
      else
      begin
         Children.Add(attr.ShortName);
      end;
   end;
   Result := 0;
end;


constructor TFatNode.Create;
begin
   inherited;

   h := nil;
end;

destructor TFatNode.Destroy;
begin
   CloseData;

   inherited;
end;

function TFatNode.HasChildren : Boolean;
begin
   Result := False;

   if Name = '/' then
   begin
      Result := True;
   end
   else if (Attrib and FILE_TYPE_DIR) = FILE_TYPE_DIR then
   begin
      Result := True;
   end;
end;

function TFatNode.HasData : Boolean;
begin
   if (Attrib and FILE_TYPE_DIR) <> FILE_TYPE_DIR then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

function TFatNode.GetLen : U64;
begin
   Result := Size;
end;

function TFatNode.GetAttribute(Attribute : String) : String;
begin
   if Attribute = 'system' then
   begin
      Result := 'fat';
   end
{   else if Attribute = 'user' then
   begin
      Result := IntToStr(stat.sd_uid);
   end
   else if Attribute = 'group' then
   begin
      Result := IntToStr(stat.sd_gid);
   end
   else if Attribute = 'ctime' then
   begin
      Result := CTime(stat.sd_ctime);
   end
   else if Attribute = 'mtime' then
   begin
      Result := CTime(stat.sd_mtime);
   end
   else if Attribute = 'atime' then
   begin
      Result := CTime(stat.sd_atime);
   end
   else if Attribute = 'mode' then
   begin
      Result := GetModeString(stat.sd_mode);
   end}
end;

procedure TFatNode.ListChildren(Children : TStringList);
var
   fat : TFatFs;
begin
   fat := TFatFs(fs);
   fatfs_list(fat.fs, Cluster, on_dir_entry, Children);
end;

function TFatNode.OpenData(WriteAccess : Boolean) : Boolean;
var
   fat : TFatFs;
   e : errcode_t;
begin
   fat := TFatFs(fs);
   Result := False;
   if HasData then
   begin
      e := fatfs_open_file(fat.fs, Directory, Cluster, Integer(WriteAccess), @h);
      if e = 0 then
      begin
         Result := True;
      end;
   end;
end;

function TFatNode.CloseData : Boolean;
var
   fat : TFatFs;
   e : errcode_t;
begin
   fat := TFatFs(fs);
   if Assigned(h) then
   begin
      e := fatfs_close_file(fat.fs, h);
      h := nil;
   end;
   Result := True;
end;

function TFatNode.ReadData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean;
var
   fat   : TFatFs;
   e     : errcode_t;
begin
   Result := False;
   fat := TFatFs(fs);
   if not Assigned(h) then
   begin
      OpenData(False);
   end;

   if Assigned(h) then
   begin
      e := fatfs_read_file(fat.fs, h, Offset, Len, Buffer);
      if e = 0 then
      begin
         Result := True;
      end
   end;
end;

function TFatNode.GetChild(Name : String) : TVNode;
var
   fat   : TFatFs;
   Find  : TFindEntry;
   Node  : TFatNode;
begin
   Result := nil;
   fat := TFatFs(fs);

   Find.Name   := Name;
   Find.attr.cluster := $FFFFFFFF;
   fatfs_list(fat.fs, Cluster, find_dir_entry, @Find);
   if Find.attr.Cluster <> $FFFFFFFF then
   begin
      Node := TFatNode.Create;
      Node.Name      := Name;
      Node.fs        := fs;
      Node.Cluster   := Find.attr.Cluster;
      Node.Directory := Cluster;
      Node.Size      := Find.attr.Size;
      Node.Attrib    := Find.attr.attr;

      Result := Node;
   end;
end;

function TFatFs.Mount(Options : TMountOptions) : TVNode;
var
   e     : errcode_t;
   Flags : Integer;
begin
   Result := nil;
   e := fatfs_open(PChar(v.GetUniqueName), get_vv_io_manager, @fs);
   if e = 0 then
   begin
      Result := TFatNode.Create;
      with Result as TFatNode do
      begin
         Name   := '/';
         fs     := Self;
         Cluster := 0;
         Directory := 0;
         // stat
//         e := reiserfs_stat_key(self.fs, @Key, @stat);
      end;
      if e = 0 then
      begin
         Result.Open;
      end
      else
      begin
         Log('Error ' + IntToStr(e));
         Result.Free;
         Result := nil;
      end;
   end
   else
   begin
      Log('Error ' + IntToStr(e));
   end;
end;

function TFatFs.UMount : Boolean;
var
   e : errcode_t;
begin
   Result := False;
   //Flush;
   e := fatfs_close(fs);
   if e = 0 then
   begin
      Result := True;
   end
   else
   begin
      Log('Error ' + IntToStr(e));
   end;
end;

// if we find any reiser, add an alias with the volume name and uuid
function BuildFatMaps(VolumeList : TList; Options : TStringList) : Boolean;
var
   i : Integer;
   v : TVirtualVolume;

   Buffer   : String;
   Magic    : String;
   UUID     : String;
   Name     : String;
   Offset   : Integer;
begin
   Result := False;

   SetLength(Buffer, 512);

   for i := 0 to VolumeList.Count - 1 do
   begin
      v := TVirtualVolume(VolumeList[i]);

      if v.GetLayer(Facility) = LEVEL_NONE then
      begin

         // we don't want to process RAID member disks...
         if v.GetLayer('raid') = LEVEL_QUERIED then
         begin
            v.SetLayer(Facility, LEVEL_QUERIED);
            Magic := '';
            if v.Read(0, Length(Buffer), PChar(Buffer)) then
            begin
               if (Ord(PChar(Buffer)[510]) = $55) and
                  (Ord(PChar(Buffer)[511]) = $AA) then
               begin
                  // That just makes it DOS, but it might just be a partition table
                  // MS sucks

                  Offset := 38;
                  Magic := Copy(Buffer, Offset + 16 + 1, 8);
                  if (Magic[1] = #0) and
                     (Magic[2] = #0) and
                     (Magic[3] = #0) and
                     (Magic[4] = #0) and
                     (Magic[5] = #0) and
                     (Magic[6] = #0) and
                     (Magic[7] = #0) and
                     (Magic[8] = #0) then
                  begin
                     Offset := 66;
                     Magic := Copy(Buffer, Offset + 16 + 1, 8);
                  end;

                  if (Copy(Magic, 1, 3) = 'FAT') and
                     (Ord(PChar(Buffer)[Offset]) = 41) then
                  begin
                     Log('FAT Found. magic = ' + Magic, Facility, LOG_VERBOSE);

                     UUID := IntToHex(PU16(PChar(Buffer) + Offset + 3)^, 4) + '-' +
                             IntToHex(PU16(PChar(Buffer) + Offset + 1)^, 4);

                     v.SetLayer(Facility, LEVEL_PROCESSED);
                     v.AddName('FS=fat');
                     v.AddName('UUID=' + UUID);
                     Name := Trim(Copy(Buffer, Offset + 6, 11));
                     if Length(Name) > 0 then
                     begin
                        v.AddName('LABEL=' + Name);
                     end;
                  end;
               end;
            end;
         end
         else if v.GetLayer('raid') = LEVEL_NONE then
         begin
            Log('Not scanning unchecked raid member ' + v.GetName, Facility, LOG_LOOP);
            Result := True;
         end
         else if v.GetLayer('raid') = LEVEL_PROCESSED then
         begin
            Log('Not scanning unused raid member ' + v.GetName, Facility, LOG_LOOP);
         end
         else if v.GetLayer('raid') = LEVEL_ACTIVATED then
         begin
            Log('Not scanning used raid member ' + v.GetName, Facility, LOG_LOOP);
         end
         else
         begin
            Log('Not scanning unknown status raid ' + v.GetName + ' at level ' + IntToStr(v.GetLayer('raid')), Facility, LOG_VERBOSE);
         end;
      end;
   end;
end;

initialization
   RegisterVolumeScan(Facility, BuildFatMaps);
   RegisterFileSystem('fat', TFatFs);


end.
