unit vfs_reiserfs3;

interface

uses types, vfs, libreiser3, classes, vv_io;

type
   TReiserFs = class(TVVFs)
   private
      WriteAccess : Boolean;
   public
      fs       : PReiserFS;
      function Mount(Options : TMountOptions) : TVNode; override;
      function UMount : Boolean; override;

   end;

   TReiserNode = class(TVNode)
   public
      constructor Create; override;
      destructor Destroy; override;

      function HasChildren : Boolean; override;
      function HasData : Boolean; override;
      function GetLen : U64; override;
      function GetAttribute(Attribute : String) : String; override;


      procedure ListChildren(Children : TStringList); override;
      {function Mkdir(Name : String) : Boolean; override;
      function CreateFile(Name : String) : Boolean; override;
      function Delete(Name : String) : Boolean; override;}


      function OpenData(WriteAccess : Boolean) : Boolean; override;
      function CloseData : Boolean; override;
      function ReadData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean; override;
{      function WriteData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean; override;
      function SetLen(Len : Int64) : Boolean; override;}

   protected

      key   : REISERFS_CPU_KEY;
      stat  : REISERFS_STAT2;

      function GetChild(Name : String) : TVNode; override;

   end;

implementation

uses windows, sysutils, debug, unicode, time, vv, ls;

const Facility = 'vfs_reiserfs3';

type TFindEntry = record
   Name  : String;
   Key   : REISERFS_CPU_KEY;
end;
PFindEntry = ^TFindEntry;

function find_dir_entry(name : PChar; Key : PREISERFS_CPU_KEY; priv_data : Pointer) : Integer; cdecl;
var
   Find  : PFindEntry;
begin
   Result := 0;
   Find := PFindEntry(priv_data);

   if name = Find.Name then
   begin
      begin
         CopyMemory(@Find.Key, Key, sizeof(REISERFS_CPU_KEY));
         Result := 0; // can't abort yet
      end;
   end;
end;

function on_dir_entry(name : PChar; Key : PREISERFS_CPU_KEY; priv_data : Pointer) : Integer; cdecl;
var
   Children : TStringList;
begin
   Children := TStringList(priv_data);
   if Assigned(Children) then
   begin
      Children.Add(name);
   end;
   Result := 0;
end;


constructor TReiserNode.Create;
begin
   inherited;

end;

destructor TReiserNode.Destroy;
begin

   inherited;
end;

function TReiserNode.HasChildren : Boolean;
begin
   if (stat.sd_mode and S_IFMT) = S_IFDIR then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

function TReiserNode.HasData : Boolean;
begin
   if (stat.sd_mode and S_IFMT) = S_IFREG then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

function TReiserNode.GetLen : U64;
begin
   Result := stat.sd_size;
end;

function TReiserNode.GetAttribute(Attribute : String) : String;
begin
   if Attribute = 'system' then
   begin
      Result := 'reiserfs';
   end
   else if Attribute = 'user' then
   begin
      Result := IntToStr(stat.sd_uid);
   end
   else if Attribute = 'group' then
   begin
      Result := IntToStr(stat.sd_gid);
   end
   else if Attribute = 'ctime' then
   begin
      Result := CTime(stat.sd_ctime);
   end
   else if Attribute = 'mtime' then
   begin
      Result := CTime(stat.sd_mtime);
   end
   else if Attribute = 'atime' then
   begin
      Result := CTime(stat.sd_atime);
   end
   else if Attribute = 'mode' then
   begin
      Result := GetModeString(stat.sd_mode);
   end
end;

procedure TReiserNode.ListChildren(Children : TStringList);
var
   reiser : TReiserFs;
begin
   reiser := TReiserFs(fs);
   reiserfs_list_key(reiser.fs, @Key, on_dir_entry, Children);
end;

function TReiserNode.OpenData(WriteAccess : Boolean) : Boolean;
begin
   if HasData then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

function TReiserNode.CloseData : Boolean;
begin
   if HasData then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

function TReiserNode.ReadData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean;
var
   reiser   : TReiserFs;
   Info     : REISERFS_READ_INFO;
begin
   reiser := TReiserFs(fs);
   Info.Offset := Offset;
   Info.Length := Len;
   Info.Buffer := Buffer;
   Info.Complete := 0;

   reiserfs_read_key(reiser.fs, @Key, @Info);
   Result := True;
end;

function TReiserNode.GetChild(Name : String) : TVNode;
var
   reiser : TReiserFs;
   e     : errcode_t;
   Find  : TFindEntry;
   Node  : TReiserNode;
begin
   Result := nil;
   reiser := TReiserFs(fs);

   Find.Name   := Name;
   Find.Key.k_objectid := 0;
   reiserfs_list_key(reiser.fs, @Key, find_dir_entry, @Find);
   if Find.Key.k_objectid > 0 then
   begin
      Node := TReiserNode.Create;
      Node.Name   := Name;
      Node.fs     := fs;
      Node.Key    := Find.key;

      e := reiserfs_stat_key(reiser.fs, @Node.Key, @Node.stat);
      if e = 0 then
      begin
         Result := Node;
      end
      else
      begin
         Log('Error ' + IntToStr(e));
         Node.Free;
      end;
   end;
end;

function TReiserFs.Mount(Options : TMountOptions) : TVNode;
var
   e     : errcode_t;
   Flags : Integer;
begin
   Result := nil;
   e := reiserfs_open(PChar(v.GetUniqueName), get_vv_io_manager, @fs);
   if e = 0 then
   begin
      Result := TReiserNode.Create;
      with Result as TReiserNode do
      begin
         Name   := '/';
         fs     := Self;
         Key.k_dir_id   := 1; // 1,2,1,500 is the root inode
         Key.k_objectid := 2;
         Key.k_offset   := 1;
         key.k_type     := 500;

         // stat
         e := reiserfs_stat_key(self.fs, @Key, @stat);
      end;
      if e = 0 then
      begin
         Result.Open;
      end
      else
      begin
         Log('Error ' + IntToStr(e));
         Result.Free;
         Result := nil;
      end;
   end
   else
   begin
      Log('Error ' + IntToStr(e));
   end;
end;

function TReiserFs.UMount : Boolean;
var
   e : errcode_t;
begin
   Result := False;
   //Flush;
   e := reiserfs_close(fs);
   if e = 0 then
   begin
      Result := True;
   end
   else
   begin
      Log('Error ' + IntToStr(e));
   end;
end;

// if we find any reiser, add an alias with the volume name and uuid
function BuildReiserMaps(VolumeList : TList; Options : TStringList) : Boolean;
var
   i : Integer;
   v : TVirtualVolume;

   Buffer   : String;
   Magic    : String;
   UUID     : String;
   Name     : String;
begin
   Result := False;

   SetLength(Buffer, 512);

   for i := 0 to VolumeList.Count - 1 do
   begin
      v := TVirtualVolume(VolumeList[i]);

      if v.GetLayer(Facility) = LEVEL_NONE then
      begin

         // we don't want to process RAID member disks...
         if v.GetLayer('raid') = LEVEL_QUERIED then
         begin
            v.SetLayer(Facility, LEVEL_QUERIED);
            Magic := '';
            if v.Read(65536, Length(Buffer), PChar(Buffer)) then
            begin
               if Copy(Buffer, 53, 8) = 'ReIsErFs' then
               begin
                  Magic := Copy(Buffer, 53, 8);
                  Log('ReiserFS 3.5 Found. magic = ' + Magic, Facility, LOG_VERBOSE);
               end
               else if Copy(Buffer, 53, 9) = 'ReIsEr2Fs' then
               begin
                  Magic := Copy(Buffer, 53, 9);
                  Log('ReiserFS 3.6 Found. magic = ' + Magic, Facility, LOG_VERBOSE);
               end
               else if Copy(Buffer, 53, 9) = 'ReIsEr3Fs' then
               begin
                  Magic := Copy(Buffer, 53, 9);
                  Log('ReiserFS JR Found. magic = ' + Magic, Facility, LOG_VERBOSE);
               end
               else if Copy(Buffer, 0, 7) = 'ReIsEr4' then
               begin
                  Magic := Copy(Buffer, 0, 7);
                  Log('ReiserFS 4.0 (libreiser) Found. magic = ' + Magic, Facility, LOG_VERBOSE);
               end;

               if Magic <> '' then
               begin
                  UUID := UuidToStr(Copy(Buffer, 85, 16));
                  Name := Trim(Copy(Buffer, 101, 16));

                  v.SetLayer(Facility, LEVEL_PROCESSED);
                  v.AddName('FS=reiserfs');
                  v.AddName('UUID=' + UUID);
                  if Length(Name) > 0 then
                  begin
                     v.AddName('LABEL=' + Name);
                  end;
               end;

            end;
         end
         else if v.GetLayer('raid') = LEVEL_NONE then
         begin
            Log('Not scanning unchecked raid member ' + v.GetName, Facility, LOG_LOOP);
            Result := True;
         end
         else if v.GetLayer('raid') = LEVEL_PROCESSED then
         begin
            Log('Not scanning unused raid member ' + v.GetName, Facility, LOG_LOOP);
         end
         else if v.GetLayer('raid') = LEVEL_ACTIVATED then
         begin
            Log('Not scanning used raid member ' + v.GetName, Facility, LOG_LOOP);
         end
         else
         begin
            Log('Not scanning unknown status raid ' + v.GetName + ' at level ' + IntToStr(v.GetLayer('raid')), Facility, LOG_VERBOSE);
         end;
      end;
   end;
end;


initialization
   RegisterVolumeScan(Facility, BuildReiserMaps);
   RegisterFileSystem('reiserfs', TReiserFs);

end.
