unit vfs_win32;

interface
uses types, vfs, classes, winbinfile;

type
   TWin32Fs = class(TFs)
   private
      Win32Path : String;
   public
      function Mount(Options : TMountOptions) : TVNode; override;
      function UMount : Boolean; override;

      function IsWritable : Boolean; override;

      procedure AssignPath(Path : String);
   end;

   TWin32Node = class(TVNode)
   private
      Path : String;
      BinFile : TBinaryFile;
   public
      constructor Create; override;
      destructor Destroy; override;

      function HasChildren : Boolean; override;
      function HasData : Boolean; override;
      function GetLen : U64; override;

      procedure ListChildren(Children : TStringList); override;

      function Mkdir(Name : String) : Boolean; override;
      function CreateFile(Name : String) : Boolean; override;
      function Delete(Name : String) : Boolean; override;

      function OpenData(WriteAccess : Boolean) : Boolean; override;
      function CloseData : Boolean; override;
      function ReadData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean; override;
      function WriteData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean; override;
      function SetLen(Len : Int64) : Boolean; override;

   protected
      function GetChild(Name : String) : TVNode; override;


   end;

implementation

uses windows, sysutils, debug, unicode, filectrl, vv, vv_dummy;

const Facility = 'win32';

function StatFile(Path : String; var F : TWIN32FindData) : Boolean;
var
   h : THandle;
begin
   h := Windows.FindFirstFile(PChar(Path), F);
   if h <> INVALID_HANDLE_VALUE then
   begin
      Windows.FindClose(h);
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

constructor TWin32Node.Create;
begin
   inherited;

   BinFile := nil;
end;

destructor TWin32Node.Destroy;
begin
   CloseData;

   inherited;
end;

function TWin32Node.HasChildren : Boolean;
begin
   Result := DirectoryExists(Path);
end;

function TWin32Node.HasData : Boolean;
begin
   Result := not HasChildren;
end;

function TWin32Node.GetLen : U64;
var
   F : TWIN32FindData;
   Size : Large_Integer;
begin
   Result := 0;
   if StatFile(Path, F) then
   begin
      Size.LowPart := f.nFileSizeLow;
      Size.HighPart := f.nFileSizeHigh;
      Result := Size.QuadPart;
   end
   else
   begin
      Result := 0; // This is mostlikely '..' which is kind of virtual anyway
   end;
end;

function TWin32Node.GetChild(Name : String) : TVNode;
var
   Search : String;
   F : TWIN32FindData;
begin
   Result := nil;

   if Path[Length(Path)] = '\' then
   begin
      Search := Path + Name;
   end
   else
   begin
      Search := Path + '\' + Name;
   end;

   if (Name = '.') or (Name = '..') or StatFile(Search, F) then
   begin
      Result := TWin32Node.Create;
      Result.fs := fs;
      TWin32Node(Result).Path := Search;
      Result.Name := Name;
   end;
end;

procedure TWin32Node.ListChildren(Children : TStringList);
var
   i : Integer;
   F : TSearchRec;
   Search : String;
begin
   if Path[Length(Path)] = '\' then
   begin
      Search := Path + '*.*';
   end
   else
   begin
      Search := Path + '\*.*';
   end;
   i := FindFirst(Search, faAnyFile, F);
   if i = 0 then
   begin
      while i = 0 do
      begin
         Children.Add(F.Name);
         i := FindNext(F);
      end;
      FindClose(F);
   end;
end;

function TWin32Node.Mkdir(Name : String) : Boolean;
begin
   Result := CreateDir(Path + '\' + Name);
end;

function TWin32Node.CreateFile(Name : String) : Boolean;
var
   NewFile : TBinaryFile;
begin
   NewFile := TBinaryFile.Create;
   NewFile.Assign(Path + '\' + Name);
   Result := NewFile.CreateNew;
   NewFile.Free;
end;

function TWin32Node.Delete(Name : String) : Boolean;
begin
   Result := DeleteFile(Path + '\' + Name);
end;

function TWin32Node.OpenData(WriteAccess : Boolean) : Boolean;
begin
   Result := False;

   if Assigned(binfile) then
   begin
      // is the mode compatible?
      Result := True;
   end
   else
   begin
      BinFile := TBinaryFile.Create;
      BinFile.Assign(Path);

      if WriteAccess then
      begin
         if BinFile.Open(OPEN_READ_WRITE) then
         begin
            Result := True;
         end;
      end
      else
      begin
         if BinFile.Open(OPEN_READ_ONLY) then
         begin
            Result := True;
         end;
      end;
   end;
end;

function TWin32Node.CloseData : Boolean;
begin
   if Assigned(BinFile) then
   begin
      BinFile.Free;
      BinFile := nil;
   end;
   Result := True;
end;

function TWin32Node.ReadData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean;
var
   Actual : DWORD;
begin
   Result := False;

   OpenData(False);

   BinFile.Seek(Offset);
   Actual := BinFile.BlockRead2(Buffer, Len);
   if Actual = Len then
   begin
      Result := True;
   end;
end;

function TWin32Node.WriteData(Offset : Int64; Len : Int64; Buffer : Pointer) : Boolean;
var
   Actual : DWORD;
begin
   Result := False;

   OpenData(True);

   BinFile.Seek(Offset);
   Actual := BinFile.BlockWrite2(Buffer, Len);
   if Actual = Len then
   begin
      Result := True;
   end;
end;

function TWin32Node.SetLen(Len : Int64) : Boolean;
begin
   OpenData(True);

   Result := BinFile.TruncateTo(Len);
end;


function TWin32Fs.Mount(Options : TMountOptions) : TVNode;
begin
   Result := nil;
   Win32Path := Options.GetParameter(0);
   if DirectoryExists(Win32Path) then
   begin
      Result := TWin32Node.Create;
      TWin32Node(Result).path := Win32Path;
      Result.fs := self;
      Result.Open;
   end
   else
   begin
      Log('Win32 Path not found ' + Win32Path);
   end;
end;

function TWin32Fs.UMount : Boolean;
begin
   Result := False;
end;

function TWin32Fs.IsWritable : Boolean;
begin
   Result := True;
end;


procedure TWin32Fs.AssignPath(Path : String);
begin
   Win32Path := Path;
end;

function BuildDriveMaps(VolumeList : TList; Options : TStringList) : Boolean;
var
   Drive : Char;
   DriveString : String;
   F     : TWIN32FindData;
   v     : TVirtualVolume;
   KnownDrives : String;
   i : Integer;
begin
   Result := False;

   for i := 0 to VolumeList.Count - 1 do
   begin
      v := TVirtualVolume(VolumeList[i]);

      if v.GetLayer(Facility) = LEVEL_ACTIVATED then
      begin
         KnownDrives := KnownDrives + Copy(v.GetName, 1, 1);
      end;
   end;


   // these are not volume based so we don't care about volume list...
   // we need a dummy volume to represent the drive letters...

   for Drive := 'a' to 'z' do
   begin
      if Pos(Drive, KnownDrives) = 0 then
      begin
         DriveString := Drive + ':\*.*';

         if StatFile(DriveString, F) then
         begin
            Log(DriveString + ' exists', Facility);
            v := TDummyVolume.Create;
            v.AddName(Drive + ':');
            v.AddName('FS=' + Facility);
            v.SetLayer(Facility, LEVEL_ACTIVATED);
            VolumeList.Add(v);
         end
      end;
   end;
end;


initialization
   RegisterFileSystem(Facility, TWin32Fs);
   RegisterVolumeScan(Facility, BuildDriveMaps);

end.
