unit vmdk;

interface

uses classes, sysutils, vfs, debug, strings, vv, types;


   type sparse_extent_header = packed record
      magic             : array[1..4] of char;
      version           : u32;
      flags             : u32;
      capacity          : u64;
      grainSize         : u64;
      descriptorOffset  : u64;
      descriptorSize    : u64;
      numGTEsPerGT      : u32;
      rgdOffset         : u64;
      gdOffset          : u64;
      overHead          : u64;
      uncleanShutdown   : u8;
      binarymagic       : array[1..4] of char;
      padding           : array[1..435] of char;
   end;

   type gte = array[0..511] of U32;
   type Pgte = ^gte;

   Psparse_extent_header = ^sparse_extent_header;

  // A vmware vmdk grain access volume
  type TGrainVolume = class(TBoundryVolume)
  private
     Node       : TVNode;
     header     : sparse_extent_header;
     gd_buffer  : String; // The entire grain directory is kept in memory
     gd         : PU32;
     gt         : gte; // the current grain table
     CurrentGT  : U32; // Only one grain table is kept in memory at a time
     GrainSize  : U64;
     GTCoverage : U64;

     procedure LoadGT(gde : U32);
  public
    function BoundryRead(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;
    function BoundryWrite(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;

    function Open(WriteAccess : Boolean) : Boolean; override;
    function Close : Boolean; override;

    procedure SetNode(Node : TVNode);

    function GetLength : Int64; override;
    function IsWritable : Boolean; override;

  end;


function LoadVMDKDescriptor(F : TVNode; Offset : Int64; Len : Int64; Dir : TVNode) : TVirtualVolume;
function FindVMDKDescriptor(Dir : TVNode; F : TVNode) : TVirtualVolume;

implementation

uses vv_map, vv_loop, windows;

const Facility = 'vmdk';

procedure TGrainVolume.LoadGT(gde : U32);
var
   offset : U64;
begin
   if CurrentGT <> gde then
   begin
      Log('Loading gde ' + IntToStr(gde), Facility);
      Offset := PU32(PChar(gd) + (gde * sizeof(U32)))^;
      Log('Offset = sector ' + IntToStr(Offset), Facility);
      Offset := Offset * 512;
      Log('Offset = ' + IntToStr(Offset), Facility);
      Node.ReadData(Offset, sizeof(gte), @gt);
      CurrentGT := gde;
   end;
end;

function TGrainVolume.BoundryRead(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean;
var
   gde : U32;
   gte : U32;
   data_offset : U64;
begin
   Log('Grain read at offset ' + IntToStr(Offset), Facility);
   gde := Offset div GTCoverage;
   Log('Looking in gt number ' + IntToStr(gde), Facility);
   LoadGT(gde);
   gte := (Offset mod GTCoverage) div GrainSize;
   Log('Looking at grain ' + IntToStr(gte), Facility);
   Log(IntToStr(gt[gte]), Facility);

   data_offset := Offset mod GrainSize;
   if gt[gte] > 0 then
   begin
      Node.ReadData(gt[gte] * 512 + data_offset, Length, Buffer);
   end
   else
   begin
      // read from the parent is NYI
      FillMemory(Buffer, Length, 0);
   end;

   Result := True;
end;

function TGrainVolume.BoundryWrite(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean;
begin
   Result := False;
end;

function TGrainVolume.Open(WriteAccess : Boolean) : Boolean;
begin
   if WriteAccess then
   begin
      Result := False;
   end
   else
   begin
      Result := Node.OpenData(WriteAccess);
      if Result then
      begin
         Node.ReadData(0, sizeof(sparse_extent_header), @header);
         Log('sparse extent version ' + IntToStr(header.version), Facility);
         Log('There are ' + IntToStr(header.numGTEsPerGT) + ' entries');
         GrainSize := header.grainSize * 512;
         Log('The grain size is ' + IntToStr(grainSize));

         SetBoundry(GrainSize);

         // Load the GD
         SetLength(gd_buffer, sizeof(U32) * header.numGTEsPerGT);
         Node.ReadData(header.gdOffset * 512, Length(gd_buffer), PChar(gd_buffer));
         gd := PU32(PChar(gd_buffer));

         GTCoverage := header.numGTEsPerGT * GrainSize;
         CurrentGT := $ffffffff;
      end;
   end;
end;

function TGrainVolume.Close : Boolean;
begin
   Result := False;
end;

procedure TGrainVolume.SetNode(Node : TVNode);
begin
   self.Node := Node;
end;

function TGrainVolume.GetLength : Int64;
begin
   Result := Header.capacity * 512;
end;

function TGrainVolume.IsWritable : Boolean;
begin
   Result := False;
end;


function FindVMDKDescriptor(Dir : TVNode; F : TVNode) : TVirtualVolume;
var
   Magic    : String;
   Header   : sparse_extent_header;
begin
   Result := nil;
   SetLength(Magic, 4);
   F.ReadData(0, 4, PChar(Magic));
   if Magic = 'KDMV' then
   begin
      F.ReadData(0, sizeof(sparse_extent_header), @Header);
      Result := LoadVMDKDescriptor(F, Header.descriptorOffset * 512, Header.descriptorSize * 512, Dir);
   end
   else if Magic = 'COWD' then
   begin
   end
   else
   begin
      // Assume this is just the descriptor?
      Result := LoadVMDKDescriptor(F, 0, F.GetLen, Dir);
   end;

end;

function LoadVMDKDescriptor(F : TVNode; Offset : Int64; Len : Int64; Dir : TVNode) : TVirtualVolume;
var
   Buffer : String;
   Descriptor : TStringList;
   i : Integer;
   s : String;
   NextOffset : Int64;
   v     : TMapVolume;
   Ext   : TStringList;

   Version : String;
   CID    : String;
   PCID   : String;
   CType  : String;
   Extent : String;
   Abort : Boolean;

   procedure ReadExtent(E : String);
   var
      Size : Int64;
      Node : TVNode;
      loop  : TLoopVolume;
      grain : TGrainVolume;
      next : TVirtualVolume;
   begin
      Ext := TStringList.Create;
      Tokenise(E, Ext);
      if Ext.Count >= 4 then
      begin
         if Ext[2] = 'FLAT' then
         begin
            Log('We can use a loop for ' + Ext[3], Facility);
         end
         else if Ext[2] = 'SPARSE' then
         begin
            Log('We will use the grain based vv', Facility);
         end
         else
         begin
            Log('Unsupported extent', Facility);
            Abort := True;
         end;

         if not Abort then
         begin
            // try and open the file
            Node := dir.FindNode(Ext[3]);
            if Assigned(Node) then
            begin
               Next := nil;

               if Ext[2] = 'FLAT' then
               begin
                  // make a loop device...
                  loop := TLoopVolume.Create;
                  loop.AddName(Ext[3]);
                  loop.SetNode(Node);
                  next := loop;
               end
               else if Ext[2] = 'SPARSE' then
               begin
                  grain := TGrainVolume.Create;
                  grain.AddName(Ext[3]);
                  grain.SetNode(Node);
                  next := grain;
               end;

               if Assigned(next) then
               begin
                  next.Open(False);
                  Size := StrToInt64(Ext[1]) * 512;
                  v.AddMap(NextOffset, Size, next, 0);
                  NextOffset := NextOffset + Size;
               end
               else
               begin
                  Abort := True;
               end;
            end
            else
            begin
               Abort := True;
            end;
         end;
      end
      else
      begin
         Abort := True;
      end;
      Ext.Free;
      Log(E, Facility);
   end;
begin
   Result := nil;

   SetLength(Buffer, Len);
   F.ReadData(Offset, Len, PChar(Buffer));

   Abort := False;
   v := TMapVolume.Create;
   v.AddName(F.Name);
   NextOffset := 0;

   Descriptor := TStringList.Create;
   Descriptor.Text := Buffer;
   for i := 0 to Descriptor.Count - 1 do
   begin
      S := Trim(Descriptor[i]);
      if Length(S) = 0 then
      begin
         // skip
      end
      else if S[1] = '#' then
      begin
         // skip
      end
      else
      begin
         if StartsWith(S, 'version=', Version) then
         begin
            Log('Version ' + Version, Facility);
            if Version <> '1' then
            begin
               Log('Only version 1 is supported', Facility, LOG_ERROR);
               Abort := True;
            end;
         end
         else if StartsWith(S, 'CID=', CID) then
         begin
            Log('CID  = ' + CID, Facility);
         end
         else if StartsWith(S, 'parentCID=', PCID) then
         begin
            Log('PCID = ' + PCID, Facility);
            if PCID <> 'ffffffff' then
            begin
               Log('Delta disks are not supported', Facility, LOG_ERROR);
               Abort := True;
            end;
         end
         else if StartsWith(S, 'createType=', CType) then
         begin
            Log('Type = ' + CType, Facility);
         end
         else if StartsWith(S, 'RW ', Extent) then
         begin
            ReadExtent(S);
         end
         else if StartsWith(S, 'RDONLY ', Extent) then
         begin
            ReadExtent(S);
         end
         else if StartsWith(S, 'NOACCESS ', Extent) then
         begin
            ReadExtent(S);
         end;
         //Log(S);
      end;
   end;
   Descriptor.Free;

   if Abort then
   begin
      v.Free;
   end
   else
   begin
      if v.Open(False) then
      begin
         Result := v;
      end
      else
      begin
         v.Free;
      end;
   end;
end;

end.
