unit vv;

interface

uses windows, classes, winbinfile, types;

const
   LEVEL_NONE        = 0;
   LEVEL_QUERIED     = 1;
   LEVEL_PROCESSED   = 2;
   LEVEL_ACTIVATED   = 3;

{
   if a volume has not been queried then LEVEL_NONE will be returned
   Once it has been queried, LEVEL_QUERIED is assigned.  If it is found
   to not to be a constituant part then it is left at QUERIED.
   If it is a constituant then it will be set to LEVEL_PROCESSED.
   Once all the constituants are found, they are all set to ACTIVATED

   It is possible for a volume to be partly ACTIVATED, in the case of LVM
   where one PV has enough for one of it's LV then it will be ACTIVATED.
   The scan will then have to examine ACTIVATED and PROCESSED volumes
   to activate any outstanding LV

   In the case of RAID, it is missleading to look at a volume which has
   not yet been examined for RAID (because the metadata is as the end of the
   volume as a false detection at the start is possible).  In this case, a scan
   should ignore volumes which are not in the 'QUERIED' state, this will ensure
   only non-raid disks are examined.
}

type
  // Todo: ref counting if we are the target of a map
  TVirtualVolume = class(TObject)
  protected
    Names   : TStringList;
    Path    : String;
    //Abstraction : Integer; // this is no good, we need a stack of abstraction layers?
    Layers  : TStringList;
  public
    constructor Create; virtual;
    destructor Destroy; override;

    function Read(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; virtual; abstract;
    function Write(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; virtual; abstract;

    function Open(WriteAccess : Boolean) : Boolean; virtual; abstract;
    function Close : Boolean; virtual; abstract;

    function GetName : String;
    procedure GetAllNames(Names : TStringList);
    function GetUniqueName : String;
    procedure AddName(Name : String);
    procedure RemoveName(Name : String);
    function MatchName(Name : String) : Boolean;
    function GetNameTag(Name : String) : String;

    function GetLength : Int64; virtual;
    function IsWritable : Boolean; virtual;

    function GetLayer(Layer : String) : Integer;
    procedure SetLayer(Layer : String; Level : Integer);


  end;


  // An unalligned access filter
  TAlignedVolume = class(TVirtualVolume)
  protected
    Alignment : Integer;
  public
    constructor Create; override;

    function Read(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;
    function Write(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;
    function AlignedRead(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; virtual; abstract;
    function AlignedWrite(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; virtual; abstract;

    procedure SetAlign(Size : Integer);
  end;

  // A boundry based access filter
  // reads/writes may not cross a boundry.  If they do they are
  // split into multiple reads/writes
  TBoundryVolume = class(TVirtualVolume)
  protected
    Boundry : Integer;
  public
    constructor Create; override;

    function Read(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;
    function Write(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;
    function BoundryRead(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; virtual; abstract;
    function BoundryWrite(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; virtual; abstract;

    procedure SetBoundry(Size : Integer);
  end;

  procedure BuildVolumeList(VolumeList : TList);
  procedure ScanVolumeList(VolumeList : TList);

type
   TVolumeScan = function (VolumeList : TList; Options : TStringList) : Boolean;

   procedure RegisterVolumeScan(Name : String; Func : TVolumeScan);


implementation

uses sysutils, debug, strings, native, lvm2, dmraid, winioctl, vv_map,
     vv_win95, vv_win32, winver;

const Facility = 'vv';

var
//   OSis95 : Boolean;
   VolumeScan : TStringList;



////////////////////////////////////////////////////////////////////////////////

constructor TVirtualVolume.Create;
begin
   inherited;
   Log('TVirtualVolume.Create', Facility, LOG_VERBOSE);
   Names := TStringList.Create;
   Layers := TStringList.Create;
end;

destructor TVirtualVolume.Destroy;
begin
   Log('TVirtualVolume.Destroy', Facility, LOG_VERBOSE);
   Names.Free;
   Layers.Free;
   inherited;
end;

function TVirtualVolume.GetName : String;
begin
   if Names.Count > 0 then
   begin
      Result := Names[0];
   end
   else
   begin
      Result := 'Unnamed';
   end;
end;

procedure TVirtualVolume.GetAllNames(Names : TStringList);
begin
   if Assigned(Names) then
   begin
      Names.Text := self.Names.Text;
   end;
end;

function TVirtualVolume.GetUniqueName : String;
begin
   Result := '0x' + IntToHex(Integer(self), 8);
end;

procedure TVirtualVolume.AddName(Name : String);
begin
   Names.Add(Name);
end;

procedure TVirtualVolume.RemoveName(Name : String);
var
   i : Integer;
begin
   i := Names.IndexOf(Name);
   if i >= 0 then
   begin
      Names.Delete(i);
   end;
end;

function TVirtualVolume.MatchName(Name : String) : Boolean;
var
   i : Integer;
begin
   Result := False;

   if Name = GetUniqueName then
   begin
      Result := True;
   end
   else
   begin
      i := Names.IndexOf(Name);
      if i >= 0 then
      begin
         Result := True;
      end
   end;
end;

// if you pass in 'FS=' it will return the value of the FS
// or '' if there is none
function TVirtualVolume.GetNameTag(Name : String) : String;
begin
   Result := Names.Values[Name];
end;

function TVirtualVolume.GetLength : Int64;
begin
   Result := 0;
end;

function TVirtualVolume.IsWritable : Boolean;
begin
   Result := False;
end;

function TVirtualVolume.GetLayer(Layer : String) : Integer;
var
   i : Integer;
begin
   i := Layers.IndexOf(Layer);
   if i >= 0 then
   begin
      Result := Integer(Layers.Objects[i]);
   end
   else
   begin
      Result := LEVEL_NONE;
   end;
end;

procedure TVirtualVolume.SetLayer(Layer : String; Level : Integer);
var
   i : Integer;
begin
   i := Layers.IndexOf(Layer);
   if i >= 0 then
   begin
      Layers.Objects[i] := TObject(Level);
   end
   else
   begin
      Layers.AddObject(Layer, TObject(Level));
   end;
end;


// Alligned IO could be made into a class
// which wraps the underlying volume
// but this can be our test code

constructor TAlignedVolume.Create;
begin
   inherited;

   Alignment := 512; // default
end;

// any sanity checks?
procedure TAlignedVolume.SetAlign(Size : Integer);
begin
   Alignment := Size;
end;

function TAlignedVolume.Read(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean;
var
   Head        : Integer;
   Tail        : Integer;
   Temp        : String;
   TempLen     : Integer;
begin
   Log('TAlignedVolume.Read ' + IntToStr(Offset) + ' len ' + IntToStr(Length), Facility, LOG_VERBOSE);

   if Alignment = 0 then
   begin
      Result := AlignedRead(Offset, Length, Buffer);
   end
   else
   begin
      Head := Offset mod Alignment;
      Tail := Length mod Alignment;

      if (Head = 0) and (Tail = 0) then
      begin
         Result := AlignedRead(Offset, Length, Buffer);
      end
      else
      begin
         // we got work to do
         if Head <> 0 then
         begin
            Log('TAlignedVolume.Read Alligning offset', Facility, LOG_ERROR);
            SetLength(Temp, Alignment);
            Result := AlignedRead(Offset - Head, Alignment, PChar(Temp));
            if not Result then exit;
            TempLen := Alignment - Head;
            if TempLen > Length then
            begin
               TempLen := Length;
            end;
            CopyMemory(Buffer, PChar(Temp) + Head, TempLen);

            Buffer := PChar(Buffer) + TempLen;
            Offset := Offset + TempLen;
            Length := Length - TempLen;

            Tail := Length mod Alignment;
         end;

         if Tail = 0 then
         begin
            Result := AlignedRead(Offset, Length, Buffer);
         end
         else
         begin
            Log('TAlignedVolume.Read Alligning length', Facility, LOG_ERROR);
            Result := AlignedRead(Offset, Length - Tail, Buffer);
            if not Result then exit;

            SetLength(Temp, Alignment);
            Result := AlignedRead(Offset + Length - Tail, Alignment, PChar(Temp));
            if not Result then exit;

            CopyMemory(PChar(Buffer) + Length - Tail, PChar(Temp), Tail);
         end;
      end;
   end;
end;


// to write a partial block we need to first read the full block and
// then copy over some of the data and write the full block back
function TAlignedVolume.Write(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean;
var
   Head        : Integer;
   Tail        : Integer;
   Temp        : String;
   TempLen     : Integer;
begin
   Log('TAlignedVolume.Write ' + IntToStr(Offset) + ' len ' + IntToStr(Length), Facility, LOG_VERBOSE);
   if Alignment = 0 then
   begin
      Result := AlignedWrite(Offset, Length, Buffer);
   end
   else
   begin
      Head := Offset mod Alignment;
      Tail := Length mod Alignment;

      if (Head = 0) and (Tail = 0) then
      begin
         Result := AlignedWrite(Offset, Length, Buffer);
      end
      else
      begin
         // we got work to do
         if Head <> 0 then
         begin
            Log('TAlignedVolume.Write Alligning offset', Facility, LOG_ERROR);
            SetLength(Temp, Alignment);
            Result := AlignedRead(Offset - Head, Alignment, PChar(Temp));
            if not Result then exit;

            TempLen := Alignment - Head;
            if TempLen > Length then
            begin
               TempLen := Length;
            end;

            CopyMemory(PChar(Temp) + Head, Buffer, TempLen);
            Result := AlignedWrite(Offset - Head, Alignment, PChar(Temp));
            if not Result then exit;

            Buffer := PChar(Buffer) + TempLen;
            Offset := Offset + TempLen;
            Length := Length - TempLen;

            Tail := Length mod Alignment;
         end;

         if Tail = 0 then
         begin
            Result := AlignedWrite(Offset, Length, Buffer);
         end
         else
         begin
            Log('TAlignedVolume.Write Alligning length', Facility, LOG_ERROR);
            Result := AlignedWrite(Offset, Length - Tail, Buffer);
            if not Result then exit;

            SetLength(Temp, Alignment);
            Result := AlignedRead(Offset + Length - Tail, Alignment, PChar(Temp));
            if not Result then exit;

            CopyMemory(PChar(Temp), PChar(Buffer) + Length - Tail, Tail);
            Result := AlignedWrite(Offset + Length - Tail, Alignment, PChar(Temp));

         end;
      end;
   end;
end;


constructor TBoundryVolume.Create;
begin
   inherited;
   
   Boundry := 0;
end;

function TBoundryVolume.Read(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean;
var
   Start  : Int64;
   Actual : Int64;
begin
   if Boundry = 0 then
   begin
      Result := BoundryRead(Offset, Length, Buffer);
   end
   else
   begin
      Result := True;
      while Result and (Length > 0) do
      begin
         Start := Offset mod Boundry;
         if Length <= Boundry - Start then
         begin
            Actual := Length;
         end
         else
         begin
            Log('TBoundryVolume.Read being split', Facility, LOG_VERBOSE);
            Actual := Boundry - Start;
         end;
         Result := BoundryRead(Offset, Actual, Buffer);
         Offset := Offset + Actual;
         Length := Length - Actual;
         Buffer := PChar(Buffer) + Actual;
      end;
   end;
end;

function TBoundryVolume.Write(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean;
var
   Start  : Int64;
   Actual : Int64;
begin
   if Boundry = 0 then
   begin
      Result := BoundryWrite(Offset, Length, Buffer);
   end
   else
   begin
      Result := True;
      while Result and (Length > 0) do
      begin
         Start := Offset mod Boundry;
         if Length <= Boundry - Start then
         begin
            Actual := Length;
         end
         else
         begin
            Log('TBoundryVolume.Write being split', Facility, LOG_VERBOSE);
            Actual := Boundry - Start;
         end;
         Result := BoundryWrite(Offset, Actual, Buffer);
         Offset := Offset + Actual;
         Length := Length - Actual;
         Buffer := PChar(Buffer) + Actual;
      end;
   end;
end;

procedure TBoundryVolume.SetBoundry(Size : Integer);
begin
   Boundry := Size;
end;


// can this get registered as a scan type?
procedure BuildVolumeList(VolumeList : TList);
begin
   ScanVolumeList(VolumeList);
end;


procedure RegisterVolumeScan(Name : String; Func : TVolumeScan);
begin
   if not Assigned(VolumeScan) then
   begin
      VolumeScan := TStringList.Create;
   end;
   VolumeScan.AddObject(Name, @Func);
end;

procedure ScanVolumeList(VolumeList : TList);
var
   Modified : Boolean;
   i : Integer;
   F : TVolumeScan;
   Options : TStringList;
begin
   Options := TStringList.Create;
   Options.AddObject('removable', TObject(1));
   if Assigned(VolumeScan) then
   begin
      Modified := True;

      while Modified do
      begin
         Modified := False;

         for i := 0 to VolumeScan.Count - 1 do
         begin
            Log('Executing ' + VolumeScan[i] + ' scan', Facility, LOG_VERBOSE);
            F := TVolumeScan(VolumeScan.Objects[i]);
            if F(VolumeList, Options) then
            begin
               Modified := True;
            end;
         end;
      end;
   end;
end;


end.
