unit vv_common;

interface

uses classes, vfs, netlib;

var
   VolumeList : TList;
   pwd : TVNode;
   root : TVNode;


procedure Setup;
function process_command(cmd : String) : Boolean;
function ShellMount(Pwd : TVNode; Parameters : TStringList) : TVNode;

type TDaemonContext = class
public
   pwd      : TVNode;
   Endpoint : TNetEndpoint;
   Handles  : TList;
   Shutdown : Boolean;

   constructor Create;
   procedure SendString(S : String);
   procedure SendError(S : String);
   procedure StartBinary;
   procedure SendBinary(Data : PChar; Len : Integer);
   procedure EndBinary;
   function RecvString : String;
   procedure RecvBinary(Data : PChar; Len : Integer);
end;

type TShellCommand = function (Pwd : TVNode; Parameters : TStringList) : TVNode;
type TDaemonCommand = function (Context : TDaemonContext) : Boolean;

implementation

uses Debug, strings, vv, vv_io, vv_loop, sysutils, unicode, vfs_ext2,
     types, vv_map, libext2fs, vv_win32, windows, winsock, libreiser3, md5,
     version, libfat, vmdk;

var
   ShellCommands : TStringList;
   DaemonCommands : TStringList;


procedure RegisterShellCommand(Name : String; Func : TShellCommand);
begin
   if not assigned(ShellCommands) then
   begin
      ShellCommands := TStringList.Create;
   end;
   ShellCommands.AddObject(Name, @Func);
end;

procedure RegisterDaemonCommand(Name : String; Func : TDaemonCommand);
begin
   if not assigned(DaemonCommands) then
   begin
      DaemonCommands := TStringList.Create;
   end;
   DaemonCommands.AddObject(Name, @Func);
end;

function ShellHelp(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   i : Integer;
begin
   Result := Pwd;
   Log('Available commands:');
   for i := 0 to ShellCommands.Count - 1 do
   begin
      Log(ShellCommands[i]);
   end;
   Log('Please see http://www.chrysocome.net/virtualvolumes for up to date documentation');
end;

// basic setup script
// perhaps this belongs in vfs?
function DaemonExit(Context : TDaemonContext) : Boolean;
begin
   Context.Shutdown := False;
   Result := False;
end;

function DaemonShutdown(Context : TDaemonContext) : Boolean;
begin
   Context.Shutdown := True;
   Result := False;
end;

function DaemonLs(Context : TDaemonContext) : Boolean;
var
   Cmd      : String;
   Node     : TVNode;
   Child    : TVNode;
   Children : TStringList;
   i        : Integer;
begin
   Context.SendString('What dir?');
   Cmd := Context.RecvString;
   Node := Context.pwd.FindNode(Cmd);
   if Assigned(Node) then
   begin
      Children := TStringList.Create;
      Children.Sorted := True;
      Node.ListChildren(Children);
      for i := 0 to Children.Count - 1 do
      begin
         Child := Node.GetChildNode(Children[i]);
         if Assigned(Child) then
         begin
            if Assigned(Child.MountPoint) then
            begin
               Context.SendString('T=M');
               Context.SendString('X=' + Child.GetAttribute('system'));
            end
            else if Child.HasChildren then
            begin
               Context.SendString('T=D');
            end
            else
            begin
               Context.SendString('T=F');
               Context.SendString('S=' + IntToStr(Child.GetLen));
            end;
            Context.SendString('U=' + Child.GetAttribute('user'));
            Context.SendString('G=' + Child.GetAttribute('user'));
            Context.SendString('C=' + Child.GetAttribute('ctime'));
            Context.SendString('M=' + Child.GetAttribute('mtime'));
            Context.SendString('A=' + Child.GetAttribute('atime'));
            Context.SendString('F=' + Child.GetAttribute('mode'));

            Context.SendString('N=' + Child.Name);
            Child.Close;
         end
         else
         begin
            Log('Does ' + Children[i] + ' exist?');
         end;
      end;
      Children.Free;

      Context.SendError('Done');
      Node.Close;
   end
   else
   begin
      Context.SendError('Not found');
   end;

   Result := True;
end;

function DaemonGet(Context : TDaemonContext) : Boolean;
var
   Cmd      : String;
   Node     : TVNode;
   Buffer   : String;
   Offset   : Int64;
   Size     : Int64;
begin
   Context.SendString('What file?');
   Cmd := Context.RecvString;
   Node := Context.pwd.FindNode(Cmd);
   if Assigned(Node) then
   begin
      Context.StartBinary;
      SetLength(Buffer, 4096);
      Size := Node.GetLen;

      Offset := 0;
      while Size > 0 do
      begin
         if Size < Length(Buffer) then
         begin
            SetLength(Buffer, Size);
         end;
         if not Node.ReadData(Offset, Length(Buffer), PChar(Buffer)) then
         begin
            break;
         end;
         Context.SendBinary(PChar(Buffer), Length(Buffer));
         Offset := Offset + Length(Buffer);
         Size := Size - Length(Buffer);
      end;
      Context.EndBinary;
      Context.SendError('Done');

      Node.Close;
   end
   else
   begin
      Context.SendError('Not found');
   end;

   Result := True;
end;

function DaemonOpen(Context : TDaemonContext) : Boolean;
var
   Cmd   : String;
   Node  : TVNode;
   i     : Integer;
begin
   Context.SendString('What file?');
   Cmd := Context.RecvString;
   Node := Context.pwd.FindNode(Cmd);
   if Assigned(Node) then
   begin
      i := Context.Handles.IndexOf(nil);
      if i < 0 then
      begin
         Context.Handles.Add(Node);
         i := Context.Handles.Count - 1;
      end
      else
      begin
         Context.Handles[i] := Node;
      end;
      Context.sendString('H=' + IntToStr(i));
      Context.SendError('Done');
   end
   else
   begin
      Context.SendError('Not found');
   end;
   Result := True;
end;

function DaemonCreate(Context : TDaemonContext) : Boolean;
var
   Cmd   : String;
   Node  : TVNode;
   Path  : String;
   Name  : String;
   p : Integer;
begin
   Context.SendString('What file?');
   Cmd := Context.RecvString;
   Log(Cmd);
   p := rpos('/', Cmd);
   if p > 0 then
   begin
      Path := Copy(Cmd, 1, p - 1);
      Name := Copy(Cmd, p + 1, Length(Cmd));
   end
   else
   begin
      Path := '.';
      Name := Cmd;
   end;
   Log('Creating file ' + Name);
   Log('In path ' + Path);

   Node := Context.pwd.FindNode(Path);
   if Assigned(Node) then
   begin
      if Node.CreateFile(Name) then
      begin
         Context.SendError('Done');
      end
      else
      begin
         Context.SendError('Error');
      end;
   end
   else
   begin
      Context.SendError('Path not found');
   end;
   Result := True;
end;

function DaemonRead(Context : TDaemonContext) : Boolean;
var
   SHandle : String;
   SOffset : String;
   SLength : String;
   Node : TVNode;
   i : Integer;
   Offset : Int64;
   Size   : Int64;
   Buffer : String;
begin
   Context.SendString('What handle?');
   SHandle := Context.RecvString;
   //Log('Handle = ' + SHandle);
   SOffset := Context.RecvString;
   //Log('Offset = ' + SOffset);
   SLength := Context.RecvString;
   //Log('Length = ' + SLength);

   Node := nil;
   i := StrToInt(SHandle);
   if (i >= 0) and (i < Context.Handles.Count) then
   begin
      Node := Context.Handles[i];
   end;

   if Assigned(Node) then
   begin
      Offset := StrToInt(SOffset);
      Size   := StrToInt(SLength);
      SetLength(Buffer, Size);
      if Node.ReadData(Offset, Length(Buffer), PChar(Buffer)) then
      begin
         Context.StartBinary;
         Context.SendBinary(PChar(Buffer), Length(Buffer));
         Context.EndBinary;
         Context.SendError('Done');
      end
      else
      begin
         Context.SendError('Error');
      end;
   end
   else
   begin
      Context.SendError('Not found');
   end;
   Result := True;
end;

function DaemonWrite(Context : TDaemonContext) : Boolean;
var
   SHandle : String;
   SOffset : String;
   SLength : String;
   Node : TVNode;
   i : Integer;
   Offset : Int64;
   Size   : Int64;
   Buffer : String;
begin
   Context.SendString('What handle?');
   SHandle := Context.RecvString;
   SOffset := Context.RecvString;
   SLength := Context.RecvString;

   Node := nil;
   i := StrToInt(SHandle);
   if (i >= 0) and (i < Context.Handles.Count) then
   begin
      Node := Context.Handles[i];
   end;

   if Assigned(Node) then
   begin
      Offset := StrToInt(SOffset);
      Size   := StrToInt(SLength);
      SetLength(Buffer, Size);
      Context.RecvBinary(PChar(Buffer), Size);

      if Node.WriteData(Offset, Length(Buffer), PChar(Buffer)) then
      begin
         Context.SendError('Done');
      end
      else
      begin
         Context.SendError('Error');
      end;
   end
   else
   begin
      Context.SendError('Not found');
   end;
   Result := True;
end;

function DaemonClose(Context : TDaemonContext) : Boolean;
var
   SHandle : String;
   Node : TVNode;
   i : Integer;
begin
   Context.SendString('What handle?');
   SHandle := Context.RecvString;

   Node := nil;
   i := StrToInt(SHandle);
   if (i >= 0) and (i < Context.Handles.Count) then
   begin
      Node := Context.Handles[i];
   end;

   if Assigned(Node) then
   begin
      Node.Close;
      Context.Handles[i] := nil;
      Context.SendError('Done');
   end
   else
   begin
      Context.SendError('Not found');
   end;


   Result := True;
end;

function DaemonShow(Context : TDaemonContext) : Boolean;
var
   Filter   : String;
   Names    : TStringList;
   i        : Integer;
   v        : TVirtualVolume;
   n : Integer;
begin
   Context.SendString('What type?');
   Filter := Context.RecvString;

   Names := TStringList.Create;
   try
      for i := 0 to VolumeList.Count - 1 do
      begin
         v := TVirtualVolume(VolumeList[i]);
         if not v.MatchName(Filter) then
         begin
            continue;
         end;
         Context.SendString('V=' + v.GetName);
         if v.IsWritable then
         begin
            Context.SendString('M=rw');
         end
         else
         begin
            Context.SendString('M=ro');
         end;
         Context.SendString('S=' + IntToStr(v.GetLength));
         v.GetAllNames(Names);
         for n := 1 to Names.Count - 1 do
         begin
            Context.SendString('A=' + Names[n]);
         end;
      end;
      Context.SendError('Done');
   finally
      Names.Free;
   end;
   Result := True;
end;

function DaemonMount(Context : TDaemonContext) : Boolean;
var
   Path : String;
   Node : TVNode;
   Options : String;
   Parameters : TStringList;
begin
   Context.SendString('What path?');
   Path := Context.RecvString;
   Options := Context.RecvString;

   Node := Context.Pwd.FindNode(Path);
   if Assigned(Node) then
   begin
      Parameters := TStringList.Create;
      Tokenise('mount ' + Options, Parameters);
      ShellMount(Node, Parameters);
      Parameters.Free;


      if Assigned(Node.overlay) then
      begin
         if not Assigned(Node.parent) then
         begin
            pwd := Node.overlay;
            Root := pwd; // chroot ?
            Context.pwd := pwd;
         end;
      end;

      Node.Close;
      Context.SendError('Done');
   end
   else
   begin
      Context.SendError('Not found');
   end;

   Result := True;
end;

function DaemonLoad(Context : TDaemonContext) : Boolean;
var
   v     : TLoopVolume;
   Path : String;
   Node : TVNode;
begin
   Context.SendString('What path?');
   Path := Context.RecvString;

   Node := Context.Pwd.FindNode(Path);
   if Assigned(Node) then
   begin
      v := TLoopVolume.Create;
      v.AddName(Path);
      v.SetNode(Node);
      if v.Open(Node.IsWritable) then
      begin
         VolumeList.Add(v);
         ScanVolumeList(VolumeList);
      end;

//      Node.Close;
      Context.SendError('Done');
   end
   else
   begin
      Context.SendError('Not found');
   end;

   Result := True;
end;

constructor TDaemonContext.Create;
begin
   Handles := TList.Create;
   Shutdown := False;
end;

procedure TDaemonContext.SendString(S : String);
begin
   Endpoint.SendStringN(S);
end;

procedure TDaemonContext.SendError(S : String);
begin
   Endpoint.SendStringN('E=' + S);
end;

procedure TDaemonContext.StartBinary;
begin
   SendError('Binary');
end;

procedure TDaemonContext.SendBinary(Data : PChar; Len : Integer);
begin
   Endpoint.SendInteger(Len);
   Endpoint.SendBuffer(Data, Len);
end;

procedure TDaemonContext.EndBinary;
begin
   Endpoint.SendInteger(0);
end;

function TDaemonContext.RecvString : String;
begin
   Result := Endpoint.RecvStringN;
end;

procedure TDaemonContext.RecvBinary(Data : PChar; Len : Integer);
begin
   // how will the client enter binary mode?
   Endpoint.RecvBuffer(Data, Len);
end;


function ShellDaemon(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   Server   : TNetServer;
   S        : TSocket;
   Cmd      : String;

   Context : TDaemonContext;
   i : Integer;
   DaemonResult : Boolean;

begin
   Result := Pwd;

   Server := TNetServer.Create(9000);
   Server.Enable(nil);

   Context := TDaemonContext.Create;
   Context.pwd := pwd;

   while not Context.Shutdown do
   begin
      S := Server.DoAccept;
      Context.Endpoint := TNetEndpoint.CreateServer(S);
      try
         Context.SendString('Virtual Volumes Server version ' + AppVersion);
         DaemonResult := True;

         while DaemonResult do
         begin
            Cmd := Context.RecvString;
            Log(Cmd, 'server', LOG_VERBOSE);

            i := DaemonCommands.IndexOf(Cmd);
            if i >= 0 then
            begin
               DaemonResult := TDaemonCommand(DaemonCommands.Objects[i])(Context);
            end
            else
            begin
               Context.SendError('Unknown command');
            end;
         end;

      except
         on E : TSocketException do
         begin
            Log(E.Message, 'server', LOG_ERROR);
         end;
      end;
      Context.Endpoint.Shutdown(SD_BOTH);
      Context.Endpoint.Free;
   end;
   Server.Free;
end;

function ShellLs(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   Children : TStringList;
   i        : Integer;
   Child    : TVNode;
   Line     : String;
begin
   Result := pwd;

   Children := TStringList.Create;
   Children.Sorted := True;
   pwd.ListChildren(Children);
   for i := 0 to Children.Count - 1 do
   begin
      Child := Pwd.GetChildNode(Children[i]);
      if Assigned(Child) then
      begin
         if Child.HasChildren then
         begin
            Line := 'd ';
         end
         else
         begin
            Line := '  ';
         end;
         Line := Line + Format('%8d %s', [Child.GetLen, Encode(Children[i])]);
         Log(Line);
         Child.Close;
      end
      else
      begin
         Log('Does ' + Children[i] + ' exist?');
      end;
   end;
   Children.Free;
end;

function ShellCd(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   Child : TVNode;
   Name  : String;
   Temp  : TVNode;
begin
   Result := Pwd;

   if Parameters.Count = 1 then
   begin
      while Result.HasParent do
      begin
         Temp := Result.GetParentNode;
         Result.Close;
         Result := Temp;
      end;
   end
   else if Parameters.Count = 2 then
   begin

      Name := Parameters[1];

      if Name = '.' then
      begin
      end
      else if Name = '..' then
      begin
         Result := Pwd.GetParentNode;
         Pwd.Close;
      end
      else
      begin
         Child := Pwd.GetChildNode(Name);
         if Assigned(Child) then
         begin
            if Child.HasChildren then
            begin
               Result := Child;
               Pwd.Close;
            end
            else
            begin
               Log('Not a directory');
               Child.Close;
            end;
         end
         else
         begin
            Log('No such file or directory');
         end;
      end;
   end
   else
   begin
      Log('Wrong number of parameters');
   end;
end;

function ShellCat(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   Child    : TVNode;
   Offset   : Int64;
   Size     : Int64;
   Buffer   : String;
   i        : Integer;
   Name     : String;
begin
   Result := Pwd;
   for i := 1 to Parameters.Count - 1 do
   begin
      Name := Parameters[i];

      Child := Pwd.GetChildNode(Name);
      if Assigned(Child) then
      begin
         Child.OpenData(False);
         SetLength(Buffer, 4096);
         Size := Child.GetLen;

         Offset := 0;
         while Size > 0 do
         begin
            if Size < Length(Buffer) then
            begin
               SetLength(Buffer, Size);
            end;
            if not Child.ReadData(Offset, Length(Buffer), PChar(Buffer)) then
            begin
               break;
            end;
            Log(Buffer);
            Offset := Offset + Length(Buffer);
            Size := Size - Length(Buffer);
         end;

         Child.CloseData;
         Child.Close;
      end
      else
      begin
         Log('No such file or directory');
      end;
   end;
end;

// optional -n blocksize
// optional filename
// default: read md5sums and check each file in there using 4096 block
function ShellMd5(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   Child    : TVNode;
   Offset   : Int64;
   Size     : Int64;
   Buffer   : String;
   i        : Integer;
   Name     : String;
   Context  : MD5Context;
   Digest   : MD5Digest;
   BlockSize   : Integer;
   Checksum    : String;

   function CalculateChecksum(Name : String) : String;
   begin
      Result := '';
      
      Child := Pwd.GetChildNode(Name);
      if Assigned(Child) then
      begin
         if Child.HasData then
         begin
            Child.OpenData(False);
            MD5Init(Context);
            SetLength(Buffer, BlockSize);
            Size := Child.GetLen;

            Offset := 0;
            while Size > 0 do
            begin
               if Size < Length(Buffer) then
               begin
                  SetLength(Buffer, Size);
               end;
               if not Child.ReadData(Offset, Length(Buffer), PChar(Buffer)) then
               begin
                  break;
               end;
               MD5Update(Context, PChar(Buffer), Length(Buffer));
               Offset := Offset + Length(Buffer);
               Size := Size - Length(Buffer);
            end;
            MD5Final(Context, Digest);
            Result := MD5Print(Digest);
            Child.CloseData;
         end;

         Child.Close;
      end
      else
      begin
         Log('No such file or directory');
      end;
   end;
begin
   Result := Pwd;
   BlockSize := 4096;

   if Parameters.Count > 2 then
   begin
      if Parameters[1] = '-n' then
      begin
         BlockSize := StrToInt(Parameters[2]);
         Parameters.Delete(1);
         Parameters.Delete(1);
      end;
   end;

   if Parameters.Count > 1 then
   begin
      for i := 1 to Parameters.Count - 1 do
      begin
         Name := Parameters[i];
         Checksum := CalculateChecksum(Name);
         Log(Checksum + ' ' + IntToStr(BlockSize) + ' ' + Name);
      end;
   end
   else
   begin
      // load the md5sums file
      Child := Pwd.GetChildNode('md5sums');
      if Assigned(Child) then
      begin
         // read the file into a string list...
         SetLength(Buffer, Child.GetLen);
         Child.OpenData(False);
         Child.ReadData(0, Child.GetLen, PChar(Buffer));
         Parameters.Text := Buffer;
         Child.CloseData;
         Child.Close;
         for i := 1 to Parameters.Count - 1 do
         begin
            Name := Copy(Parameters[i], 35, Length(Parameters[i]));
            Checksum := CalculateChecksum(Name);
            if Length(Checksum) = 32 then
            begin
               Log(Parameters[i]);
               if Copy(Parameters[i], 1, 32) <> Checksum then
               begin
                  Log('Checksum failed. Got ' + Checksum);
                  break;
               end;
            end;
         end;
      end
      else
      begin
         Log('No md5sums file to check');
      end;
   end;

end;

function ShellMkdir(pwd : TVNode; Parameters : TStringList) : TVNode;
begin
   Result := pwd;
   if not pwd.Mkdir(Parameters[1]) then
   begin
      Log('Error creating directory');
   end;
end;


function ShellTouch(pwd : TVNode; Parameters : TStringList) : TVNode;
var
   i : Integer;
   Name : String;
begin
   Result := Pwd;

   for i := 1 to Parameters.Count - 1 do
   begin
      Name := Parameters[i];

      Pwd.CreateFile(Name);
   end;
end;

function ShellDelete(pwd : TVNode; Parameters : TStringList) : TVNode;
var
   i : Integer;
   Name : String;
begin
   Result := Pwd;

   for i := 1 to Parameters.Count - 1 do
   begin
      Name := Parameters[i];

      Pwd.Delete(Name);
   end;
end;

{procedure do_load_vv_win32(Name : String);
var
   v : TWin32Volume;
begin
   v := TWin32Volume.Create;
   v.SetPath(Name);
   if v.Open(True) then
   begin
      VolumeList.Add(v);
   end
   else
   begin
      Log('Error loading ' + Name);
      v.Free;
   end;
end;}

function ShellLoad(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   v     : TLoopVolume;
   Node  : TVNode;
   i     : Integer;
   Name  : String;
   DoScan : Boolean;
{   m     : TMapVolume;
   o : Int64;}
begin
   Result := Pwd;
   DoScan := False;
   for i := 1 to Parameters.Count - 1 do
   begin
      Name := Parameters[i];
      if Assigned(Pwd) then
      begin
         Node := pwd.FindNode(Name);
         if Assigned(Node) then
         begin
            v := TLoopVolume.Create;
            v.AddName(Name);
            v.SetNode(Node);
            if v.Open(Node.IsWritable) then
            begin
               {if Name = 'pup001' then // testing the MapVolume
               begin
                  Log('asdf');
                  // test a map volume here...
                  m := TMapVolume.Create;
                  m.AddName(v.GetName);
                  o := 0;
                  while o < v.GetLength do
                  begin
                     m.AddMap(o, 512, v, o);
                     o := o + 512;
                  end;
                  VolumeList.Add(m);
               end
               else}
               begin
                  VolumeList.Add(v);
               end;
               DoScan := True;
            end
            else
            begin
               Log('Error opening ' + Name);
               Node.Close;
               v.Free;
            end;
         end
         else
         begin
            Log('Node not found ' + Name);
         end;
      end
      else
      begin
         Log('No pwd to lookup node');
      end;
   end;
   if DoScan then
   begin
      ScanVolumeList(VolumeList);
   end;
end;

function ShellVMWareLoad(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   v     : TVirtualVolume;
   Node  : TVNode;
   i     : Integer;
   Name  : String;
   DoScan : Boolean;
   Magic : String;
begin
   Result := Pwd;
   DoScan := False;
   for i := 1 to Parameters.Count - 1 do
   begin
      Name := Parameters[i];
      if Assigned(Pwd) then
      begin
         Node := pwd.FindNode(Name);
         if Assigned(Node) then
         begin
            v := FindVMDKDescriptor(pwd, Node);
            if Assigned(v) then
            begin
               VolumeList.Add(v);
               ScanVolumeList(VolumeList);
            end;
            Node.Close;
         end
         else
         begin
            Log('Node not found ' + Name);
         end;
      end
      else
      begin
         Log('No pwd to lookup node');
      end;
   end;
   if DoScan then
   begin
      ScanVolumeList(VolumeList);
   end;
end;

function ShellStat(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   Node : TVNode;
   i : Integer;
   Name : String;
begin
   Result := Pwd;
   for i := 1 to Parameters.Count - 1 do
   begin
      Name := Parameters[i];
      Node := pwd.FindNode(Name);
      if Assigned(Node) then
      begin
         Log('Found node ' + Name);
         Node.Close;
      end;
   end;
end;

procedure CopyData(FromNode : TVNode; ToNode : TVNode);
var
   Offset   : Int64;
   Size     : Int64;
   Buffer   : String;
begin
   SetLength(Buffer, 4096);
   Size := FromNode.GetLen;

   Offset := 0;
   while Size > 0 do
   begin
      if Size < Length(Buffer) then
      begin
         SetLength(Buffer, Size);
      end;
      if not FromNode.ReadData(Offset, Length(Buffer), PChar(Buffer)) then
      begin
         Log('Error reading data');
         break;
      end;
      if not ToNode.WriteData(Offset, Length(Buffer), PChar(Buffer)) then
      begin
         Log('Error writing data');
         break;
      end;
      Offset := Offset + Length(Buffer);
      Size := Size - Length(Buffer);
   end;

   if Size = 0 then
   begin
      ToNode.SetLen(Offset);
   end
end;

function ShellCopy(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   FromName : String;
   ToName   : String;
   FromNode : TVNode;
   ToNode   : TVNode;
   NewNode  : TVNode;
begin
   Result := Pwd;
   if Parameters.Count = 3 then
   begin
      FromName := Parameters[1];
      ToName   := Parameters[2];

      Log('Copy file ' + FromName + ' to ' + ToName);

      FromNode := pwd.FindNode(FromName);
      if Assigned(FromNode) then
      begin
         Log('Found from node');
         if FromNode.HasData then
         begin
            ToNode := pwd.FindNode(ToName);
            if Assigned(ToNode) then
            begin
               Log('Found to node');
               if ToNode.HasData then
               begin
                  CopyData(FromNode, ToNode);
               end
               else
               begin
                  // create a new file
                  Log('Creating file called ' + FromNode.Name);
                  if ToNode.CreateFile(FromNode.Name) then
                  begin
                     NewNode := ToNode.GetChildNode(FromNode.Name);
                     ToNode.Close;
                     ToNode := NewNode;
                     CopyData(FromNode, ToNode);
                  end;
               end;
               ToNode.Close;
            end
            else
            begin
               // create a new file in pwd
               if Pwd.CreateFile(ToName) then
               begin
                  ToNode := Pwd.GetChildNode(ToName);
                  if Assigned(ToNode) then
                  begin
                     CopyData(FromNode, ToNode);
                     ToNode.Close;
                  end;
               end;
            end;
         end
         else
         begin
            Log('FromNode is not a file');
         end;
         FromNode.Close;
      end;
   end;
end;

function ShellShow(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   i     : Integer;
   v     : TVirtualVolume;
   Names : TStringList;
   n     : Integer;
   Mode  : String;
   Filter : Boolean;
   FilterString : String;
begin
   Result := Pwd;

   if Parameters.Count = 2 then
   begin
      Filter := True;
      FilterString := Parameters[1];
   end
   else
   begin
      Filter := False;
   end;

   Names := TStringList.Create;
   try
      for i := 0 to VolumeList.Count - 1 do
      begin
         v := TVirtualVolume(VolumeList[i]);
         if Filter then
         begin
            if not v.MatchName(FilterString) then
            begin
               continue;
            end;
         end;
         if v.IsWritable then
         begin
            Mode := 'rw';
         end
         else
         begin
            Mode := 'ro';
         end;
         Log(v.GetName + ' (' + human_size(v.GetLength) + ' ' + Mode + ')');
         v.GetAllNames(Names);
         for n := 1 to Names.Count - 1 do
         begin
            Log(' alias ' + Names[n]);
         end;
         Log('');
      end;
   finally
      Names.Free;
   end;
end;

function ShellUmount(Pwd : TVNode; Parameters : TStringList) : TVNode;
begin
   Result := Pwd;
   // is this a mount point?
   Pwd.UMount;
end;

// 0     1     2            3
// mount ext2  \Device\asdf [options]
// mount win32 c:\          [options]
// mount sftp  user@host    [options]
function ShellMount(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   fs       : TFs;
   i        : Integer;
   Options  : TMountOptions;
   j        : Integer;

   procedure CopyOptions(From : Integer);
   var
      i : Integer;
   begin
      Options := TMountOptions.Create;
      for i := From to Parameters.Count - 1 do
      begin
         Options.Add(Parameters[i]);
      end;
   end;
begin
   Result := pwd;
   fs := nil;

   if Parameters.Count > 1 then
   begin
      for j := 0 to FileSystems.Count - 1 do
      begin
         if Parameters[1] = FileSystems[j] then
         begin
            // class found, instnaciate it (it's a bit of black magic)
            fs := TFsClass(FileSystems.Objects[j]).Create;
            //Log('I have created a ' + fs.ClassName);
            if fs is TVVFs then
            begin
               if Parameters.Count >= 2 then
               begin
                  for i := 0 to VolumeList.Count - 1 do
                  begin
                     if TVirtualVolume(VolumeList[i]).MatchName(Parameters[2]) then
                     begin
                        TVVFs(fs).AssignVolume(TVirtualVolume(VolumeList[i]));
                        CopyOptions(3);
                        if not pwd.Mount(fs, Options) then
                        begin
                           fs.Free;
                           fs := nil;
                        end;
                        break;
                     end;
                  end;
                  if not Assigned(fs) then
                  begin
                     Log('Virtual Volume ' + Parameters[2] + ' not found');
                  end;
               end;
            end
            else // it's not block based
            begin
               CopyOptions(2);
               if not pwd.Mount(fs, Options) then
               begin
                  fs.Free;
                  fs := nil;
               end;
            end;
            break;
         end;
      end;
   end;
end;

function ShellBlockDump(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   Name     : String;
   Offset   : Int64;
   Length   : Int64;
   Buffer   : String;

   S  : String;
   S2 : String;
   i  : Integer;
   j  : Integer;
   v        : TVirtualVolume;

begin
   Result := Pwd;
   if Parameters.Count = 4 then
   begin
      Name := Parameters[1];
      Offset := StrToInt64(Parameters[2]);
      Length := StrToInt64(Parameters[3]);

      for j := 0 to VolumeList.Count - 1 do
      begin
         v := TVirtualVolume(VolumeList[j]);
         if v.MatchName(Name) then
         begin
            SetLength(Buffer, Length);
            if v.Read(Offset, Length, PChar(Buffer)) then
            begin
               // dump the output
               S := IntToHex(Offset, 8) + ' ';
               for i := 0 to Length - 1 do
                  begin
                     S := S + IntToHex(Ord(PChar(Buffer)[i]), 2);
                     if not (ord(PChar(Buffer)[i]) in [0, 7, 8, 9]) then
                     begin
                        S2 := S2 + PChar(Buffer)[i];
                     end
                     else
                     begin
                        S2 := S2 + ' ';
                     end;
                     S := S + ' ';
                     if System.Length(S) >= 56 then
                     begin
                        Log(S + ' ' + S2);
                        S := IntToHex(i + Offset + 1, 8) + ' ';
                        s2 := '';
                     end;
                  end;
               end;
            end;
         end;
      end;
end;

// Read block numbers to test block device access
function ShellTest1(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   i        : Integer;
   Offset   : Int64;
   Buffer   : String;
   v        : TVirtualVolume;
   id       : Integer;
   Counter  : Integer;
   Name     : String;
   Error    : Boolean;
   Child    : TVNode;

   r : Integer;
   h : Pointer;
begin
   Result := Pwd;
{   if Parameters.Count = 2 then
   begin
      Name := Parameters[1];

      for i := 0 to VolumeList.Count - 1 do
      begin
         v := TVirtualVolume(VolumeList[i]);
         if v.MatchName(Name) then
         begin
            Log('Scanning ' + v.GetName);
            SetLength(Buffer, 1024);
            Offset := 0;
            Counter := 0;
            Error := False;
            while Offset < v.GetLength do
            begin
               v.Read(Offset, 1024, PChar(Buffer));
               id := PU32(PChar(Buffer))^;
               //Log(IntToStr(id));
               if id <> Counter then
               begin
                  Log('Identify failed at offset ' + IntToStr(Offset));
                  Error := True;
                  break;
               end;
               Counter := Counter + 1;
               Offset := Offset + 1024;
            end;
            if not Error then
            begin
               Log('Tested ' + IntToStr(Counter) + 'k');
            end;
         end;
      end;
   end;    }
{   r := reiserfs_open('\Device\Harddisk2\Partition1', get_vv_io_manager, @h);
   if r = 0 then
   begin
      // the partition is open...
//      reiserfs_list(h, '/');
      reiserfs_close(h);
   end;}
{   r := fatfs_open('fat32.img', get_vv_io_manager, @h);
   if r = 0 then
   begin
      fatfs_dump(h);
      fatfs_close(h);
   end;}

end;

// read big blocks which will need to be chopped up
// to test boundry io
function ShellTest4(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   i        : Integer;
   j        : Integer;
   Offset   : Int64;
   Buffer   : String;
   v        : TVirtualVolume;
   Name     : String;
begin
   Result := Pwd;
   if Parameters.Count = 2 then
   begin
      Name := Parameters[1];

      for i := 0 to VolumeList.Count - 1 do
      begin
         v := TVirtualVolume(VolumeList[i]);
         if v.GetName = Name then
         begin
            SetLength(Buffer, 4096);
            for j := 0 to 128 do
            begin
               Offset := j * 1024;
               if not v.Read(Offset, Length(Buffer), PChar(Buffer)) then
               begin
                  break;
               end;
            end;
         end;
      end;
   end;
end;

function ShellTest2(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   i : Integer;
begin
   Result := Pwd;
{   for i := 1 to 1000 do
   begin
      Pwd.Mkdir('testdir' + IntToStr(i));
      Pwd.CreateFile('testfile' + IntToStr(i));
   end;}


end;

// test aligned IO
function ShellTest3(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   i : Integer;
   offset : Integer;
   length : Integer;
   v1 : TWin32Volume;
   v2 : TWin32Volume;
   Buffer1 : String;
//   Buffer2 : String;
   Fill     : Integer;
begin
   Result := Pwd;

   v1 := TWin32Volume.Create;
   v1.SetPath('c:\temp\random.dat');
   v1.Open(True);
   v1.SetAlign(0);

   v2 := TWin32Volume.Create;
   v2.SetPath('c:\temp\random2.dat');
   v2.Open(True);
   v2.SetAlign(512);

   for i := 0 to 1024 do
   begin
      Offset := Random(1024 * 1024 * 9);
      Length := Random(65536);
      Fill := Random(256);
      Log('Test ' + IntToStr(i) + ' ' + IntToStr(Offset) + ' -> ' + IntToStr(Length) + ' = ' + IntToStr(Fill));
      SetLength(Buffer1, Length);
//      SetLength(Buffer2, Length);
//      FillMemory(PChar(Buffer1), Length, 1);
      FillMemory(PChar(Buffer1), Length, Fill);
      if v1.Write(Offset, Length, PChar(Buffer1)) then
      begin
         if v2.Write(Offset, Length, PChar(Buffer1)) then
         begin
            // compare...
{            if not CompareMem(PChar(Buffer1), PChar(Buffer2), Length) then
            begin
               Log('error at ' + IntToStr(Offset) + ' ' + IntToStr(Length));
               break;
            end;}
         end
         else
         begin
            Log('error');
            break;
         end;
      end
      else
      begin
         Log('error');
         break;
      end;
   end;
   v2.Close;
   v1.Close;
   v2.Free;
   v1.Free;
end;

function ShellSource(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   Commands : TStringList;
   i        : Integer;
   Name     : String;
begin
   if Parameters.Count >= 2 then
   begin
      if (Parameters[1] = 'win32') and (Parameters.Count = 3) then
      begin
         Name := Parameters[2];
         Commands := TStringList.Create;
         try
            try
               Commands.LoadFromFile(Name);
               for i := 0 to Commands.Count - 1 do
               begin
                  process_command(Commands[i]);
               end;
            except
               on E : EFOpenError do
               begin
                  Log(E.Message);
               end;
            end;
         finally
            Commands.Free;
         end;
      end
      else
      begin
         Log('vfs source NYI');
      end;
   end;
   Result := vv_common.Pwd;
end;

function ShellDump(Pwd : TVNode; Parameters : TStringList) : TVNode;
begin
   Result := Pwd;
   Root.DumpDir('', True);
end;

function ShellDebug(pwd : TVNode; Parameters : TStringList) : TVNode;
var
   Level : Integer;
begin
   Result := pwd;
   if Parameters.Count = 3 then
   begin
      if Parameters[2] = 'ERROR' then
      begin
         Level := LOG_ERROR;
      end
      else if Parameters[2] = 'NORMAL' then
      begin
         Level := LOG_NORMAL;
      end
      else if Parameters[2] = 'VERBOSE' then
      begin
         Level := LOG_VERBOSE;
      end
      else if Parameters[2] = 'LOOP' then
      begin
         Level := LOG_LOOP;
      end
      else if Parameters[2] = 'DUMP' then
      begin
         Level := LOG_DUMP;
      end
      else if Parameters[2] = 'MAX' then
      begin
         Level := LOG_MAX;
      end
      else
      begin
         Level := StrToIntDef(Parameters[2], LOG_NORMAL);
      end;
      SetLogLevel(Parameters[1], Level);
   end;
end;

function ShellExit(Pwd : TVNode; Parameters : TStringList) : TVNode;
begin
   Result := nil;
end;

function ShellScan(Pwd : TVNode; Parameters : TStringList) : TVNode;
begin
   Result := Pwd;
   BuildVolumeList(VolumeList);
end;

function ShellAutoMount(Pwd : TVNode; Parameters : TStringList) : TVNode;
var
   i     : Integer;
   v     : TVirtualVolume;
   Names : TStringList;
   n     : Integer;
   Mode  : String;
   Filter : Boolean;
   FilterString : String;
   Name  : String;
   DirName : String;
   Fs    : String;
   Child : TVNode;
begin
   Result := Pwd;

   if Parameters.Count = 2 then
   begin
      Filter := True;
      FilterString := Parameters[1];
   end
   else
   begin
      Filter := False;
   end;

   Names := TStringList.Create;
   try
      for i := 0 to VolumeList.Count - 1 do
      begin
         v := TVirtualVolume(VolumeList[i]);
         if Filter then
         begin
            if not v.MatchName(FilterString) then
            begin
               continue;
            end;
         end;
         {if v.IsWritable then
         begin
            Mode := 'rw';
         end
         else
         begin
            Mode := 'ro';
         end;}
         //Log(v.GetName + ' (' + human_size(v.GetLength) + ' ' + Mode + ')');
         {v.GetAllNames(Names);
         for n := 1 to Names.Count - 1 do
         begin
            Log(' alias ' + Names[n]);
         end;
         Log('');}
         Fs := v.GetNameTag('FS');
         if Length(Fs) > 0 then
         begin
            // Pick a name which is not in use
            {Name := v.GetNameTag('LABEL');
            if Length(Name) = 0 then
            begin
               Name := v.GetNameTag('UUID');
            end;
            if Length(Name) = 0 then
            begin
               Name := v.GetName;
            end;}
            Name := v.GetName;

            Child := Pwd.FindNode(Name);
            if Assigned(Child) then
            begin
               // Already exists
               Child.Close;
            end
            else
            begin
               DirName := StringReplace(Name, ' ', '_', [rfReplaceAll]);
               DirName := StringReplace(DirName, '\', '_', [rfReplaceAll]);
               DirName := StringReplace(DirName, '/', '_', [rfReplaceAll]);

               process_command('cd');
               process_command('mkdir "' + DirName + '"');
               process_command('cd "' + DirName + '"');
               process_command('mount ' + Fs + ' "' + Name + '"');
               process_command('cd');
            end;
            //process_command('dir');
         end;
      end;
   finally
      Names.Free;
   end;
end;

procedure Setup;
begin
   Log('Virtual Volume IO Driver version ' + AppVersion);
   Log(' aka Explore2fs 2.0 beta testing');
   Log('Written by John Newbigin <jn@it.swin.edu.au>');
   Log('http://www.chrysocome.net/virtualvolumes');
   Log('This program is licensed under the GPL');
   Log('');
   Log(' BACK UP YOUR DATA!');
   Log('');
   Log('This program might destroy your entire hard disk!');
   Log('');
   Log(' BACK UP YOUR DATA!');
   Log('');
   Log('If you have not done a backup, type "exit"');
   Log('');

   VolumeList := TList.Create;

   SetVolumeList(VolumeList);

   RegisterShellCommand('daemon',   ShellDaemon);
   RegisterShellCommand('server',   ShellDaemon);
   RegisterShellCommand('mkdir',    ShellMkdir);
   RegisterShellCommand('ls',       ShellLs);
   RegisterShellCommand('dir',      ShellLs);
   RegisterShellCommand('cd',       ShellCd);
   RegisterShellCommand('cat',      ShellCat);
   RegisterShellCommand('type',     ShellCat);
   RegisterShellCommand('more',     ShellCat);
   RegisterShellCommand('md5',      ShellMd5);
   RegisterShellCommand('touch',    ShellTouch);
   RegisterShellCommand('load',     ShellLoad);
   RegisterShellCommand('loadvm',   ShellVMWareLoad);
   RegisterShellCommand('stat',     ShellStat);
   RegisterShellCommand('copy',     ShellCopy);
   RegisterShellCommand('cp',       ShellCopy);
   RegisterShellCommand('umount',   ShellUmount);
   RegisterShellCommand('test',     ShellTest1);
   RegisterShellCommand('blockdump', ShellBlockDump);
   RegisterShellCommand('source',   ShellSource);
   RegisterShellCommand('help',     ShellHelp);
   RegisterShellCommand('?',        ShellHelp);
   RegisterShellCommand('dump',     ShellDump);
   RegisterShellCommand('exit',     ShellExit);
   RegisterShellCommand('quit',     ShellExit);
   RegisterShellCommand('debug',    ShellDebug);
   RegisterShellCommand('scan',     ShellScan);
   RegisterShellCommand('auto',     ShellAutoMount);
   RegisterShellCommand('delete',   ShellDelete);
   RegisterShellCommand('del',      ShellDelete);
   RegisterShellCommand('rm',       ShellDelete);
   RegisterShellCommand('unlink',   ShellDelete);
   RegisterShellCommand('show',     ShellShow);
   RegisterShellCommand('mount',    ShellMount);

   RegisterDaemonCommand('exit',    DaemonExit);
   RegisterDaemonCommand('shutdown',DaemonShutdown);
   RegisterDaemonCommand('ls',      DaemonLs);
   RegisterDaemonCommand('get',     DaemonGet);
   RegisterDaemonCommand('open',    DaemonOpen);
   RegisterDaemonCommand('create',  DaemonCreate);
   RegisterDaemonCommand('read',    DaemonRead);
   RegisterDaemonCommand('write',   DaemonWrite);
   RegisterDaemonCommand('close',   DaemonClose);
   RegisterDaemonCommand('show',    DaemonShow);
   RegisterDaemonCommand('mount',   DaemonMount);
   RegisterDaemonCommand('load',    DaemonLoad);
   RegisterDaemonCommand('loopback',DaemonLoad);
end;


function process_command(cmd : String) : Boolean;
var
   i : Integer;
   Parameters : TStringList;
   NewPwd : TVNode;
begin
   Result := True;
//Log('|' + cmd + '|');
   // chop up cmd into parameters...
   Parameters := TStringList.Create;
   if Tokenise(Cmd, Parameters) and (Parameters.Count > 0) then
   begin
      i := ShellCommands.IndexOf(Parameters[0]);
      if i >= 0 then
      begin
         NewPwd := TShellCommand(ShellCommands.Objects[i])(Pwd, Parameters);
         if Assigned(NewPwd) then
         begin
            Pwd := NewPwd;

            if Parameters[0] = 'mount' then
            begin
               if Assigned(pwd.overlay) then
               begin
                  if not Assigned(pwd.parent) then
                  begin
                     pwd := pwd.overlay;
                     Root := pwd; // chroot ?
                  end;
               end;
            end;
         end
         else
         begin
            Result := False;
         end;
      end
      else
      begin
         Log('unknown command');
      end;
   end;
   Parameters.Free;
end;


end.
