unit vv_io;

// This is an IO manager which maps the IO of libext2fs
// to a TVirtualVolume object
// the unit should be called vv_io_ext2

interface

uses windows, classes, types;

type
   errcode_t   = Longint;

   io_manager = record
    magic         : errcode_t;
    name          : PChar;
    open          : pointer;
    close         : pointer;
    set_blksize   : pointer;
    read_blk      : pointer;
    write_blk     : pointer;
    flush         : pointer;
    write_byte    : pointer;
    set_option    : pointer;
    reserved: array[1..14] of Integer;
   end;

   Pio_manager = ^io_manager;

   io_channel = record
    magic         : errcode_t;
    io_manager    : Pio_manager;
    name          : PChar;
    block_size    : Integer;
    read_error    : Pointer;
    write_error   : Pointer;
    refcount      : Integer;
    flags         : Integer;
    reserved      : array[1..14] of Integer;
    priv_data     : Pointer;
    app_data      : Pointer;
   end;
   
   Pio_channel  = ^io_channel;
   PPio_channel = ^Pio_channel;


function get_vv_io_manager : Pio_manager;
procedure SetVolumeList(VolumeList : TList);

function vv_open(name : PChar; flags : Integer; channel : PPio_channel) : errcode_t; cdecl;
function vv_close(channel : Pio_channel): errcode_t; cdecl;
function vv_set_blksize(channel : Pio_channel; blksize : Integer) : errcode_t; cdecl;
function vv_read_blk(channel : Pio_channel; block : ULong; count : Integer; data : PChar) : errcode_t; cdecl;
function vv_write_blk(channel : Pio_channel; block : ULong; count : Integer; data : PChar) : errcode_t; cdecl;
function vv_flush(channel : Pio_channel) : errcode_t; cdecl;


implementation

uses winbinfile, debug, sysutils, vv, libext2fs;

const Facility = 'vv_io';


type
   Tvv_io_manager = class(TObject)
   private
    v : TVirtualVolume;
   end;

var
   manager : io_manager;
   Volumes : TList;

procedure SetVolumeList(VolumeList : TList);
begin
   Volumes := VolumeList;
end;



function get_vv_io_manager : Pio_manager;
begin
   FillMemory(@manager, sizeof(manager), 0);

   manager.magic        := EXT2_ET_MAGIC_IO_MANAGER;
   manager.name         := 'Virtual Volume Driver';
   manager.open         := @vv_open;
   manager.close        := @vv_close;
   manager.set_blksize  := @vv_set_blksize;
   manager.read_blk     := @vv_read_blk;
   manager.write_blk    := @vv_write_blk;
   manager.flush        := @vv_flush;

   Result := @manager;
end;

function vv_open(name : PChar; flags : Integer; channel : PPio_channel) : errcode_t;
var
   vv : Tvv_io_manager;
   v : TVirtualVolume;
   i : Integer;
   Found : Boolean;
begin
   Log('vv_open ' + name, Facility);

   v := nil;

   Found := False;
   for i := 0 to Volumes.Count - 1 do
   begin
      v := TVirtualVolume(Volumes[i]);
      if (v.GetUniqueName = Name) or (v.GetName = Name) then
      begin
         Found := True;
         break;
      end;
   end;

   if Found then
   begin
      channel^ := AllocMem(sizeof(io_channel));
      FillMemory(channel^, sizeof(io_channel), 0);
      channel^.magic := EXT2_ET_MAGIC_IO_CHANNEL;
      channel^.io_manager := @manager;
      channel^.name := AllocMem(strlen(name) + 1);
      StrCopy(channel^.name, name);
      channel^.block_size := 512;
      channel^.refcount := 1;

      vv := Tvv_io_manager.Create;
      channel^.priv_data := vv;

      vv.v := v;
      Result := 0;
   end
   else
   begin
      Result := 2;
   end;
end;

function vv_close(channel : Pio_channel): errcode_t;
var
   vv : Tvv_io_manager;
begin
   Log('vv_close', Facility);
   vv := Tvv_io_manager(channel.priv_data);
   channel^.priv_data := nil;
   //vv.f.Close;
   //vv.f.free;
   //vv.f := nil;
   vv.Free;

   result := 0;
end;

function vv_set_blksize(channel : Pio_channel; blksize : Integer) : errcode_t;
begin
   Log('vv_set_blksize ' + channel^.name + ' to ' + IntToStr(blksize), Facility);
   channel.block_size := blksize;
   result := 1;
end;

function vv_read_blk(channel : Pio_channel; block : ULong; count : Integer; data : PChar) : errcode_t;
var
   vv : Tvv_io_manager;
begin
   vv := Tvv_io_manager(channel.priv_data);
   if count < 0 then
   begin
      count := -count;
      Log('vv_read_blk read ' + IntToStr(count) + ' bytes', Facility);
      if vv.v.Read(block * channel.block_size, count, data) then
      begin
         Result := 0;
      end
      else
      begin
         Result := EXT2_ET_INVALID_ARGUMENT;//GetLastError;
      end
   end
   else
   begin
      Log('vv_read_blk read ' + IntToStr(count) + ' blocks', Facility);
      if vv.v.Read(block * channel.block_size, count * channel.block_size, data) then
      begin
         Result := 0;
      end
      else
      begin
         Result := EXT2_ET_INVALID_ARGUMENT;//GetLastError;
      end
   end;
end;

function vv_write_blk(channel : Pio_channel; block : ULong; count : Integer; data : PChar) : errcode_t;
var
   vv : Tvv_io_manager;
begin
   vv := Tvv_io_manager(channel.priv_data);
   if count < 0 then
   begin
      count := -count;
      Log('vv_write_byte write ' + IntToStr(count) + ' bytes', Facility);
      if vv.v.Write(block * channel.block_size, count, data) then
      begin
         Result := 0;
      end
      else
      begin
         Log('vv_write_byte write ' + IntToStr(count) + ' bytes failed', Facility, LOG_ERROR);
         Result := EXT2_ET_INVALID_ARGUMENT;//GetLastError;
      end
   end
   else
   begin
      Log('vv_write_block write ' + IntToStr(count) + ' blocks', Facility);
      if vv.v.Write(block * channel.block_size, count * channel.block_size, data) then
      begin
         Result := 0;
      end
      else
      begin
         Log('vv_write_block write ' + IntToStr(count) + ' blocks failed', Facility, LOG_ERROR);
         Result := EXT2_ET_INVALID_ARGUMENT;//GetLastError;
      end

   end;
end;

function vv_flush(channel : Pio_channel) : errcode_t;
begin
   Log('vv_flush', Facility);
   Result := 0;
end;


end.
