unit vv_map;

interface

uses vv, classes;

type
  TMapEntry = class(TObject)
  public
    Offset : Int64;
    Length : Int64;
    Target : TVirtualVolume;
    TargetOffset : Int64;

    function InRange(fOffset : Int64; fLength : Int64) : Boolean;
    function StartInRange(fOffset : Int64; fLength : Int64) : Boolean;
    function EndInRange(fOffset : Int64; fLength : Int64) : Boolean;
  end;


  TMapVolume = class(TVirtualVolume)
  private
    Maps : TList;
    Size : Int64;
    Writable : Boolean;
  public
    constructor Create; override;
    destructor Destroy; override;

    function Read(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;
    function Write(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;

    function Open(WriteAccess : Boolean) : Boolean; override;
    function Close : Boolean; override;

    function GetLength : Int64; override;
    function IsWritable : Boolean; override;

    procedure AddMap(Offset : Int64; Length : Int64; Target : TVirtualVolume; TargetOffset : Int64);
  end;

implementation

uses debug, sysutils;

const Facility = 'vv_map';


function TMapEntry.InRange(fOffset : Int64; fLength : Int64) : Boolean;
begin
   Result := False;

   if (fOffset >= Offset) and (fOffset + fLength <= Offset + Length) then
   begin
      Result := True;
   end;
end;

function TMapEntry.StartInRange(fOffset : Int64; fLength : Int64) : Boolean;
begin
   if (fOffset >= Offset) and (fOffset < Offset + Length) then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

function TMapEntry.EndInRange(fOffset : Int64; fLength : Int64) : Boolean;
begin

   if fOffset + fLength <= Offset + Length then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

constructor TMapVolume.Create;
begin
   inherited;

   Log('TMapVolume.Create', Facility, LOG_VERBOSE);
   Maps := TList.Create;

end;

destructor TMapVolume.Destroy;
begin
   Log('TMapVolume.Destroy', Facility, LOG_VERBOSE);

   inherited;
end;


function TMapVolume.Read(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean;
var
   i        : Integer;
   Map      : TMapEntry;
   ThisLen  : Int64;
   Mapped   : Boolean;
begin
   Result := False;
   Mapped := False;
   Log('TMapVolume.Read ' + IntToStr(Offset) + ' len ' + IntToStr(Length), Facility, LOG_VERBOSE);
{   if Name = 'pup001' then
   begin
      Log(Name);
   end;}
   for i := 0 to Maps.Count - 1 do
   begin
      Map := TMapEntry(Maps[i]);

      if Map.StartInRange(Offset, Length) then
      begin
         Mapped := True;
         if Map.EndInRange(Offset, Length) then
         begin
            Result := Map.Target.Read(Map.TargetOffset - Map.Offset + Offset, Length, Buffer);
            break;
         end
         else
         begin
            // we span a map boundry, split up the read
            ThisLen := Map.Offset + Map.Length - Offset;
            Result := Map.Target.Read(Map.TargetOffset - Map.Offset + Offset, ThisLen, Buffer);
            if Result then
            begin
               Result := Read(Offset + ThisLen, Length - ThisLen, PChar(Buffer) + ThisLen);
               break;
            end;
         end;
      end;
   end;
   if not Mapped then
   begin
      Log('No map found for specified range', Facility, LOG_ERROR);
   end;
end;

// perhaps a flag which allows overlap, so we can write
// to 2 maps... ie raid 1

function TMapVolume.Write(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean;
var
   i        : Integer;
   Map      : TMapEntry;
   ThisLen  : Int64;
   Mapped   : Boolean;
begin
   Result := False;
   Mapped := False;
   Log('TMapVolume.Write ' + IntToStr(Offset) + ' len ' + IntToStr(Length), Facility);
   if Writable then
   begin
      for i := 0 to Maps.Count - 1 do
      begin
         Map := TMapEntry(Maps[i]);

         if Map.StartInRange(Offset, Length) then
         begin
            Mapped := True;
            if Map.EndInRange(Offset, Length) then
            begin
               Result := Map.Target.Write(Map.TargetOffset - Map.Offset + Offset, Length, Buffer);
               break;
            end
            else
            begin
               // we span a map boundry, split up the write
               ThisLen := Map.Offset + Map.Length - Offset;
               Result := Map.Target.Write(Map.TargetOffset - Map.Offset + Offset, ThisLen, Buffer);
               if Result then
               begin
                  Result := Write(Offset + ThisLen, Length - ThisLen, PChar(Buffer) + ThisLen);
                  break;
               end;
            end;
         end;
      end;
   end;
   if not Mapped then
   begin
      Log('No map found for specified range', Facility, LOG_ERROR);
   end;
end;

function TMapVolume.Open(WriteAccess : Boolean) : Boolean;
begin
   Log('TMapVolume.Open', Facility, LOG_VERBOSE);
   if not WriteAccess or Writable then
   begin
      Result := True;
   end
   else
   begin
      Result := False;
   end;
end;

function TMapVolume.Close : Boolean;
begin
   Result := False;
end;

// Todo: Check for overlaps with existing maps
procedure TMapVolume.AddMap(Offset : Int64; Length : Int64; Target : TVirtualVolume; TargetOffset : Int64);
var
   Map : TMapEntry;
begin
   Map := TMapEntry.Create;
   Map.Offset := Offset;
   Map.Length := Length;
   Map.Target := Target;
   Map.TargetOffset := TargetOffset;
   Maps.Add(Map);

   if Offset + Length > Size then
   begin
      Size := Offset + Length;
   end;
   if Target.IsWritable then
   begin
      Writable := True;
   end;
end;

function TMapVolume.GetLength : Int64;
begin
   Result := Size;
end;

function TMapVolume.IsWritable : Boolean;
begin
   Result := Writable;
end;

end.
