unit vv_win32;

interface
   uses vv, winbinfile, classes;

type
  TWin32Volume = class(TAlignedVolume)
  private
     f : TBinaryFile;
     WriteEnabled : Boolean;
  public
    //function Read(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;
    //function Write(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;
    function AlignedRead(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;
    function AlignedWrite(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;

    function Open(WriteAccess : Boolean) : Boolean; override;
    function Close : Boolean; override;

    procedure SetPath(Path : String); virtual;

    function GetLength : Int64; override;
    function IsWritable : Boolean; override;
  end;

  TNativeVolume = class(TWin32Volume)
  public
    function Open(WriteAccess : Boolean) : Boolean; override;
    procedure SetPath(Path : String); override;

    function GetLength : Int64; override;
  end;

function BuildVolumeListWin32(VolumeList : TList; Options : TStringList) : Boolean;
procedure BuildExtentMap(VolumeList : TList; Volume : TVirtualVolume);
  


implementation

uses debug, windows, sysutils, native, strings, vv_map, winver;

const Facility = 'vv_win32';

function TWin32Volume.AlignedRead(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean;
var
   Actual      : DWORD;
begin
   Log('TWin32.AllignedRead ' + IntToStr(Offset) + ' len ' + IntToStr(Length), Facility, LOG_VERBOSE);

   Result := False;
   if Assigned(f) then
   begin
      f.Seek(Offset);
      if f.GetPos = Offset then
      begin
         Actual := f.BlockRead2(Buffer, Length);
         if Actual = Length then
         begin
            Result := True;
         end;
      end;

   end;
end;

function TWin32Volume.AlignedWrite(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean;
var
   Actual : DWORD;
begin
   Log('TWin32.Write ' + IntToStr(Offset) + ' len ' + IntToStr(Length), Facility, LOG_VERBOSE);
   Result := False;
   if IsWritable then
   begin
      if Assigned(f) then
      begin
         f.Seek(Offset);
         if f.GetPos = Offset then
         begin
            Actual := f.BlockWrite2(Buffer, Length);
            if Actual = Length then
            begin
               Result := True;
            end
            else
            begin
               Log('TWin32.Write ' + IntToStr(Offset) + ' len ' + IntToStr(Length) + ' failed', Facility, LOG_ERROR);
            end;
         end;
      end
      else
      begin
         Log('TWin32.Write failed. File not open', Facility, LOG_ERROR);
      end;
   end
   else
   begin
      Log('TWin32.Write failed. File is read only', Facility, LOG_ERROR);
   end;
end;

function TWin32Volume.Open(WriteAccess : Boolean) : Boolean;
var
   OpenMode : Integer;
begin
   Log('TWin32Volume.Open ' + Path, Facility, LOG_VERBOSE);
   Result := False;
   Alignment := 0;

   if Assigned(f) then
   begin
      // check read/write access
      Result := True;
      exit;
   end;

   f := TBinaryFile.Create;
   f.Assign(GetName);
   if WriteAccess then
   begin
      WriteEnabled := True;
      OpenMode := OPEN_READ_WRITE;
   end
   else
   begin
      WriteEnabled := False;
      OpenMode := OPEN_READ_ONLY;
   end;
   if f.Open(OpenMode) then
   begin
      Result := True;
   end
   else
   begin
      f.Free;
      f := nil;
   end;
end;

function TWin32Volume.Close : Boolean;
begin
   if Assigned(f) then
   begin
      f.Close;
      f := nil;
   end;
   Result := True;
end;

procedure TWin32Volume.SetPath(Path : String);
begin
   Self.Path := Path;
   AddName(Path);
end;

function TWin32Volume.GetLength : Int64;
begin
   if Assigned(f) then
   begin
      Result := f.FileSize;
   end
   else
   begin
      Result := 0;
   end;
end;

function TWin32Volume.IsWritable: Boolean;
begin
   Result := WriteEnabled;
end;

////////////////////////////////////////////////////////////////////////////////

function TNativeVolume.Open(WriteAccess : Boolean) : Boolean;
var
   h : THandle;
   Mode : DWORD;
begin
   Log('TNativeVolume.Open ' + Path, Facility, LOG_VERBOSE);
   Result := False;

   if WriteAccess then
   begin
      Mode := GENERIC_READ or GENERIC_WRITE;
      WriteEnabled := True;
   end
   else
   begin
      Mode := GENERIC_READ;
      WriteEnabled := False;
   end;

   h := NTCreateFile(PChar(Path), Mode, FILE_SHARE_READ or FILE_SHARE_WRITE, nil, OPEN_EXISTING, FILE_FLAG_RANDOM_ACCESS, 0);
   if h <> INVALID_HANDLE_VALUE then
   begin
      SetAlign(GetDiskBlockSize(h));
      f := TBinaryFile.Create;
      f.AssignHandle(h);
      Result := True;
   end
end;

procedure TNativeVolume.SetPath(Path : String);
begin
   Self.Path := Path;
   AddName(Path);
end;

function TNativeVolume.GetLength : Int64;
var
   NewDevice : String;
   NewOffset : Int64;
   NewLength : Int64;
   Disk : String;
begin
   Result := 0;
   if Assigned(f) then
   begin
      if EndsWith(GetName, 'Partition0', Disk) then
      begin
         Result := GetDiskSize(f.GetHandle);
      end
      else
      begin
         if GetDiskExtents(f.GetHandle, NewDevice, NewOffset, NewLength) then
         begin
            Result := NewLength;
         end
         else
         begin
            Result := GetPartitionSize(f.GetHandle);
         end;
      end;
   end;
end;


////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


// check for duplictates and add then add the device.
function AddNativeVolume(VolumeList : TList; DeviceName : String) : Boolean;
var
   i : Integer;
   v : TNativeVolume;
begin
   Result := True;
   for i := 0 to VolumeList.Count - 1 do
   begin
      if TVirtualVolume(VolumeList[i]).GetName = DeviceName then
      begin
         Result := False;
         break;
      end;
   end;
   if Result then
   begin
      v := TNativeVolume.Create;
      v.SetPath(DeviceName);
      VolumeList.Add(v);
   end;
end;


function BuildVolumeListWin32(VolumeList : TList; Options : TStringList) : Boolean;
var
   Devices     : TStringList;
   Harddisks   : TStringList;
   Hdvolumes   : TStringList;
   i           : Integer;
   DeviceName  : String;
   Number      : String;
   DriveNo     : Integer;
   PartNo      : Integer;
   PartCount   : Integer;
   AddRemovable : Boolean;
begin
   SetErrorMode(SEM_FAILCRITICALERRORS);

   Result := False;
   AddRemovable := False;
   if Assigned(Options) then
   begin
      if Options.IndexOf('removable') >= 0 then
      begin
         AddRemovable := Boolean(Options.Objects[Options.IndexOf('removable')]);
      end;
   end;
   Devices := TStringList.Create;
   Devices.Sorted := True;
   Harddisks := TStringList.Create;
   Harddisks.Sorted := True;
   Hdvolumes := TStringList.Create;
   Hdvolumes.Sorted := True;
   NativeDir('\Device', Devices);
   for i := 0 to Devices.Count - 1 do
   begin
      DeviceName := '';

      if StartsWith(Devices[i], 'CdRom', Number) then
      begin
         DriveNo := StrToIntDef(Number, -1);
         if DriveNo >= 0 then
         begin
            DeviceName := '\Device\CdRom' + IntToStr(DriveNo);
         end;
      end
      else if StartsWith(Devices[i], 'Floppy', Number) then
      begin
         DriveNo := StrToIntDef(Number, -1);
         if DriveNo >= 0 then
         begin
            DeviceName := '\Device\Floppy' + IntToStr(DriveNo);
         end;
      end
      else if StartsWith(Devices[i], 'HarddiskVolume', Number) then
      begin
         Hdvolumes.Add('\Device\HarddiskVolume' + Number);
      end
      else if StartsWith(Devices[i], 'Harddisk', Number) then
      begin
         DriveNo := StrToIntDef(Number, -1);
         if DriveNo >= 0 then
         begin
            // scan the partitions...
            Harddisks.Add('\Device\Harddisk' + IntToStr(DriveNo));
         end;
      end;

      if Length(DeviceName) > 0 then
      begin
         Log('Native DIR found ' + DeviceName, Facility, LOG_VERBOSE);

         if AddRemovable then
         begin
            if AddNativeVolume(VolumeList, DeviceName) then
            begin
               Result := True;
            end;
         end;
      end;
   end;

   // do the hard disk partitions...
   PartCount := 0;
   for DriveNo := 0 to Harddisks.Count - 1 do
   begin
      Devices.Clear;
      NativeDir(Harddisks[DriveNo], Devices);
      for i := 0 to Devices.Count - 1 do
      begin
         if StartsWith(Devices[i], 'Partition', Number) then
         begin
            PartNo := StrToIntDef(Number, -1);
            PartCount := PartCount + 1;
            DeviceName := Harddisks[DriveNo] + '\Partition' + IntToStr(PartNo);
            Log(DeviceName, Facility, LOG_VERBOSE);
            if AddNativeVolume(VolumeList, DeviceName) then
            begin
               Result := True;
            end;
         end;
      end;
   end;

   if PartCount = 0 then
   begin
      // No partitions?  Must be Vista.
{      for i := 0 to Hdvolumes.Count - 1 do
      begin
         AddNativeVolume(VolumeList, Hdvolumes[i]);
      end;}
   end;

   Devices.Free;
   Harddisks.Free;
end;


procedure BuildExtentMap(VolumeList : TList; Volume : TVirtualVolume);
var
   v : TNativeVolume;
   p : TVirtualVolume;
   n : TMapVolume;
   NewDevice : String;
   NewOffset : Int64;
   NewLength : Int64;
   i : Integer;
begin
   if Volume is TNativeVolume then
   begin
      v := TNativeVolume(Volume);

      if GetDiskExtents(v.f.GetHandle, NewDevice, NewOffset, NewLength) then
      begin
         Log('Backdoor is ' + NewDevice, Facility, LOG_VERBOSE);
         for i := 0 to VolumeList.Count - 1 do
         begin
            p := TVirtualVolume(VolumeList[i]);
            if p.GetName = NewDevice then
            begin
               Log('The backdoor is open', Facility, LOG_VERBOSE);
               n := TMapVolume.Create;
               n.AddMap(0, NewLength, p, NewOffset);
               n.AddName(v.GetName);
               v.RemoveName(n.GetName);
               v.AddName('x' + n.GetName);
               v.SetLayer('extent', LEVEL_ACTIVATED);
               v.Close;
               VolumeList.Add(n);
               break;
            end;
         end;
      end;
   end;
end;

function BuildExtentMaps(VolumeList : TList; Options : TStringList) : Boolean;
var
   i : Integer;
   v : TVirtualVolume;
   Buffer : String;
begin
   Result := False;

   i := 0;

   while i < VolumeList.Count do
   begin
      v := TVirtualVolume(VolumeList[i]);
      if v.GetLayer('extent') = LEVEL_NONE then
      begin
         v.SetLayer('extent', LEVEL_QUERIED);
         if v.Open(True) then
         begin
           SetLength(Buffer, 512);
            if v.Read(0, 512, PChar(Buffer)) then
            begin
               Log('Native Partition Read Success', 'vv_win32', LOG_VERBOSE);
               v.SetLayer('extent', LEVEL_PROCESSED);
            end
            else
            begin
               // Build an extent map...
               BuildExtentMap(VolumeList, v);
               Result := True;
            end;
         end;
      end;
      i := i + 1;
   end;
end;

initialization
   if not OSis95 then
   begin
      RegisterVolumeScan(Facility, BuildVolumeListWin32);
      RegisterVolumeScan(Facility, BuildExtentMaps);
   end;
end.
