unit vv_win95;

interface

uses vv, diskio, classes, windows;

type
  TWin95Volume = class(TAlignedVolume)
  private
     Disk : T95Disk;
     WriteEnabled : Boolean;
  public
    constructor Create; override;
    destructor Destroy; override;

    function AlignedRead(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;
    function AlignedWrite(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean; override;

    function Open(WriteAccess : Boolean) : Boolean; override;
    function Close : Boolean; override;

    function GetLength : Int64; override;
    function IsWritable : Boolean; override;

    procedure SetDisk(DiskNumber : Integer);
  end;

{$A-,Z+}
type
 TPartEntry = record
	boot_ind    : Byte;		// 0x80 - active
	head        : Byte;		// starting head
	sector      : Byte;		// starting sector
	cyl         : Byte;		// starting cylinder
	sys_ind     : Byte;		// What partition type
	end_head    : Byte;		// end head
	end_sector  : Byte;	   // end sector
	end_cyl     : Byte;		// end cylinder
	start_sect  : UINT;	   // starting sector counting from 0
	nr_sects    : UINT;  	// nr of sectors in partition
  end;

  TDiskTable = record
   Data  : array[1..446] of byte;
   Table : array[1..4] of TPartEntry;

   Signature : WORD;
  end;

function BuildVolumeListWin95(VolumeList : TList; Options : TStringList) : Boolean;


implementation

uses debug, vv_map, sysutils, native, winioctl, winver;

const FACILITY = 'vv_win95';


constructor TWin95Volume.Create;
begin
   inherited;

   Disk := T95Disk.Create(True);
end;

destructor TWin95Volume.Destroy;
begin
   Disk.Free;

   inherited;
end;

function TWin95Volume.AlignedRead(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean;
var
   Sector : Int64;
   Count  : Int64;
   Remainder : Int64;
begin
   Result := False;

   Sector := Offset div 512;
   Remainder := Offset mod 512;
   if Remainder <> 0 then
   begin
      Log('Win95 reads must be aligned', FACILITY);
   end
   else
   begin
      Count := Length div 512;
      Remainder := Length mod 512;
      if Remainder <> 0 then
      begin
         Log('Win95 read length must be aligned', FACILITY);
      end
      else if Count > 127 then
      begin
         Log('Win95 read too large', FACILITY);
      end
      else
      begin
         Result := Disk.ReadSector(Sector, Buffer, Count);
      end;
   end;
end;

function TWin95Volume.AlignedWrite(Offset : Int64; Length : Int64; Buffer : Pointer) : Boolean;
var
   Sector : Int64;
   Count  : Int64;
   Remainder : Int64;
begin
   Result := False;
   if IsWritable then
   begin
      Sector := Offset div 512;
      Remainder := Offset mod 512;
      if Remainder <> 0 then
      begin
         Log('Win95 writes must be aligned', FACILITY);
      end
      else
      begin
         Count := Length div 512;
         Remainder := Length mod 512;
         if Remainder <> 0 then
         begin
            Log('Win95 write length must be aligned', FACILITY);
         end
         else if Count > 127 then
         begin
            Log('Win95 write too large', FACILITY);
         end
         else
         begin
            Result := Disk.WriteSector(Sector, Buffer, Count);
         end;
      end;
   end;
end;

function TWin95Volume.Open(WriteAccess : Boolean) : Boolean;
begin
   WriteEnabled := WriteAccess;
   Result := True;
end;

function TWin95Volume.Close : Boolean;
begin
   Result := True;
end;

procedure TWin95Volume.SetDisk(DiskNumber : Integer);
begin
   Disk.SetDisk(DiskNumber);
end;

function TWin95Volume.GetLength : Int64;
begin
   Result := Disk.SectorCount;
   Result := Result * 512;
end;

function TWin95Volume.IsWritable : Boolean;
begin
   Result := WriteEnabled;
end;

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////


////////////////////////////////////
////////////////////////////////////
// Windows 95 scan

// This does a Low Level scan (reads partition table from disk);
// This is the same as LL_Scan, but it uses the DiskIO unit

procedure FindWin95Part(Volume : TWin95Volume; VolumeList : TList);
var
   PartNo   : Integer;
   ExtendedStart : ULong;

   function ProcessSector(Disk : T95Disk; Sector : ULong; Depth : Integer) : Boolean;
   var
      Size     : DWORD;
      Buffer   : TDiskTable;
      i        : Integer;
      n        : TMapVolume;
      Start    : Int64;
      Len      : Int64;

   begin
      Result := True;
      Size := sizeof(Buffer);
      Log('Looking for table at sector ' + IntToStr(Sector), Facility);

      if not Disk.ReadSector(Sector, @Buffer, size div 512) then
         begin
            Result := False;
            exit;
         end;
         Log('Signature is ' + IntToHex(Buffer.Signature, 4), Facility);
         if (Buffer.Signature = $AA55) then
         begin
            for i := 1 to 4 do
            begin
               Log('Processing partitions', Facility);
               if Buffer.Table[i].sys_ind <> 0 then
               begin
                  Log('Partition           ' + IntToStr(i) + ' /dev/hdx' + IntToStr(PartNo), Facility);
                  Log(' boot indicator      ' + IntToHex(Buffer.Table[i].boot_ind,2), Facility);
                  Log(' starting head       ' + IntToStr(Buffer.Table[i].head), Facility);
                  Log(' starting sector     ' + IntToStr(Buffer.Table[i].sector), Facility);
                  Log(' starting cylinder   ' + IntToStr(Buffer.Table[i].cyl), Facility);
                  Log(' partition type      ' + IntToHex(Buffer.Table[i].sys_ind,2), Facility);
                  case Buffer.Table[i].sys_ind of
                     PARTITION_EXTENDED         : Log('Extended... (dos)', Facility);
                     PARTITION_EXTENDED_LINUX   : Log('Extended... (linux)', Facility);
                     PARTITION_EXTENDED_WIN98   : Log('Extended... (win98)', Facility);
                     PARTITION_FAT_16           : Log('FAT 16', Facility);
                     PARTITION_HUGE             : Log('VFAT', Facility);
                     PARTITION_IFS              : Log('NTFS', Facility);
                     PARTITION_LINUX            : Log('Linux', Facility);
                     PARTITION_LINUX_SWAP       : Log('Linux Swap', Facility);
                  else
                     Log('Unknown', Facility);
                  end;
                  Log('end head            ' + IntToStr(Buffer.Table[i].end_head), Facility);
                  Log('end sector          ' + IntToStr(Buffer.Table[i].end_sector), Facility);
                  Log('end cylinder        ' + IntToStr(Buffer.Table[i].end_cyl), Facility);
                  Log(' starting sector     ' + IntToStr(Buffer.Table[i].start_sect), Facility);
                  Log(' number of sectors   ' + IntToStr(Buffer.Table[i].nr_sects), Facility);
                  Log('', Facility);

                  n := TMapVolume.Create;
                  n.AddName(Volume.GetName + '/' + IntToStr(PartNo));
                  Start := Buffer.Table[i].nr_sects;
                  Start := Start * 512;
                  Len := Buffer.Table[i].start_sect + Sector;
                  Len := Len * 512;
                  n.AddMap(0, Start, Volume, Len);
                  VolumeList.Add(n);

                  Inc(PartNo);
               end;
            end;
            for i := 1 to 4 do
            begin
               if Buffer.Table[i].sys_ind in EXTENDED_PARTITIONS then
               begin
                  Log('Processing extended partition....', Facility);
                  if Depth < 32 then
                  begin
                     if PartNo < 5 then
                     begin
                        PartNo := 5;
                     end;
                     if Depth = 1 then
                     begin
                        ExtendedStart := Sector;
                     end;
                     // There is a new partition table at offset start_sect...
                     if Depth >= 2 then
                     begin
                     if not ProcessSector(Disk, ExtendedStart + Buffer.Table[i].start_sect, Depth + 1) then
                        begin
                           Log('Did not seem to be a valid extended partition', Facility);
                        end;
                     end
                     else
                     begin
                        if not ProcessSector(Disk, Sector + Buffer.Table[i].start_sect, Depth + 1) then
                        begin
                           Log('Did not seem to be a valid extended partition', Facility);
                        end;
                     end;
                  end
                  else
                  begin
                     Log('More than 32 partitions detected! This is most likely a bug!', Facility);
                  end;
               end;
            end;
         end
         else
         begin
            Result := False;
         end;
      end;
begin
   PartNo := 1;
   ExtendedStart := 0;
   ProcessSector(Volume.Disk, 0, 0);
end;

function AddWin95Volume(VolumeList : TList; DiskNumber : Integer) : Boolean;
var
   i : Integer;
   NewName : String;
   v : TWin95Volume;
begin
   Result := True;

   NewName := '$' + IntToHex(DiskNumber, 2);

   for i := 0 to VolumeList.Count -1 do
   begin
      if TVirtualVolume(VolumeList[i]).GetName = NewName then
      begin
         Result := False;
         break;
      end;
   end;

   if Result then
   begin
      v := TWin95Volume.Create;
      v.SetDisk(DiskNumber);
      v.AddName(NewName);
      VolumeList.Add(v);
      if DiskNumber >= $80 then
      begin
         FindWin95Part(v, VolumeList);
      end;
   end;
end;


function BuildVolumeListWin95(VolumeList : TList; Options : TStringList) : Boolean;
var
   DriveNo  : Integer;
   Disk     : T95Disk;
   AddRemovable : Boolean;
begin
   Result := False;

    AddRemovable := False;
   if Assigned(Options) then
   begin
      if Options.IndexOf('removable') >= 0 then
      begin
         AddRemovable := Boolean(Options.Objects[Options.IndexOf('removable')]);
      end;
   end;


   try
      Disk := T95Disk.Create(True);
   except
      on E : EFOpenError do
      begin
         exit;
      end;
   end;
   try
      DriveNo := $80;

      while true do
      begin
         Log('Scanning disk ' + IntToHex(DriveNo, 2), Facility);
         if not Disk.SetDisk(DriveNo) then
         begin
            // no more physical drives found...
            break;
         end
         else // we have the handle, lets look at the partitions....
         begin
            Log('Disk found', Facility);
            //ProcessSector(Disk, 0, 0);
            if AddWin95Volume(VolumeList, DriveNo) then
            begin
               Result := True;
            end;
            Inc(DriveNo);
         end;
      end;

      ////////////////////////////////////////////
      ////////////////////////////////////////////
      ////////////////////////////////////////////
      ////////////////////////////////////////////

      if AddRemovable then
      begin
         DriveNo := 0;

         while true do
         begin
            Log('Scanning ' + IntToHex(DriveNo, 2), Facility);
            if not Disk.SetDisk(DriveNo) then
            begin
               break;
            end
            else // we have the handle, lets look at the partitions....
            begin
               // no Partitons found, should we add the whole disk?
               if AddWin95Volume(VolumeList, DriveNo) then
               begin
                  Result := True;
               end;
               Inc(DriveNo);
            end;
         end;
      end;
   finally
      Disk.Free;
   end;
end;

initialization
   if OSis95 then
   begin
      RegisterVolumeScan(Facility, BuildVolumeListWin95);
   end;


end.
